theory Environment_Functor imports
  Main
  Applicative_Functor
begin

subsection \<open>The environment functor\<close>

type_synonym ('i, 'a) envir = "'i \<Rightarrow> 'a"

definition LIFT :: "'a \<Rightarrow> ('i, 'a) envir"
where "LIFT x = (\<lambda>_. x)"

lemma LIFT_apply [simp]: "LIFT x i = x"
by(simp add: LIFT_def)

definition ap_envir :: "('i, 'a \<Rightarrow> 'b) envir \<Rightarrow> ('i, 'a) envir \<Rightarrow> ('i, 'b) envir"
where "ap_envir f x = (\<lambda>i. f i (x i))"

adhoc_overloading Applicative_Functor.ap ap_envir

lemma ap_envir_apply [simp]: "(f \<diamond> x) i = f i (x i)"
by(simp add: ap_envir_def)

lemma af_identity_envir: "LIFT id \<diamond> x = x"
by(simp add: ap_envir_def LIFT_def)

lemma af_composition_envir: "LIFT op \<circ> \<diamond> u \<diamond> v \<diamond> w = u \<diamond> (v \<diamond> w)"
by(simp add: ap_envir_def LIFT_def)

lemma af_homomorphism_envir: "LIFT f \<diamond> LIFT x = LIFT (f x)"
by(simp add: ap_envir_def LIFT_def)

lemma af_interchange_gpv: "u \<diamond> LIFT x = LIFT (\<lambda>f. f x) \<diamond> u"
by(simp add: LIFT_def ap_envir_def)

text \<open> @{typ "('i, 'a) envir"} has combinators S,K,C \<close>
(* FIXME: add combinators *)



subsection \<open>lifting predicates to the environment functor\<close>

definition all_envir :: "('i, bool) envir \<Rightarrow> bool"
where "all_envir p \<longleftrightarrow> (\<forall>x. p x)"

lemma all_envirI [Pure.intro!, intro!]: "(\<And>x. p x) \<Longrightarrow> all_envir p"
by(simp add: all_envir_def)

lemma all_envirE [Pure.elim 2, elim]: "all_envir p \<Longrightarrow> (p x \<Longrightarrow> thesis) \<Longrightarrow> thesis"
by(simp add: all_envir_def)

lemma all_envirD: "all_envir p \<Longrightarrow> p x"
by(simp add: all_envir_def)


definition pred_envir :: "('a \<Rightarrow> bool) \<Rightarrow> ('i, 'a) envir \<Rightarrow> bool"
where "pred_envir p f = all_envir (LIFT p \<diamond> f)"

lemma pred_envir_conv: "pred_envir p f \<longleftrightarrow> (\<forall>x. p (f x))"
by(auto simp add: pred_envir_def)

lemma pred_envirI [Pure.intro!, intro!]: "(\<And>x. p (f x)) \<Longrightarrow> pred_envir p f"
by(auto simp add: pred_envir_def)

lemma pred_envirD: "pred_envir p f \<Longrightarrow> p (f x)"
by(auto simp add: pred_envir_def)

lemma pred_envirE [Pure.elim 2, elim]: "pred_envir p f \<Longrightarrow> (p (f x) \<Longrightarrow> thesis) \<Longrightarrow> thesis"
by(simp add: pred_envir_conv)

lemma pred_envir_mono: "\<lbrakk> pred_envir p f; \<And>x. p (f x) \<Longrightarrow> q (g x) \<rbrakk> \<Longrightarrow> pred_envir q g"
by blast

definition rel_envir :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('i, 'a) envir \<Rightarrow> ('i, 'b) envir \<Rightarrow> bool"
where "rel_envir p f g \<longleftrightarrow> all_envir (LIFT p \<diamond> f \<diamond> g)"

lemma rel_envir_conv: "rel_envir p f g \<longleftrightarrow> (\<forall>x. p (f x) (g x))"
by(auto simp add: rel_envir_def)

lemma rel_envir_conv_rel_fun: "rel_envir = rel_fun op ="
by(simp add: rel_envir_conv rel_fun_def fun_eq_iff)

lemma rel_envirI [Pure.intro!, intro!]: "(\<And>x. p (f x) (g x)) \<Longrightarrow> rel_envir p f g"
by(auto simp add: rel_envir_def)

lemma rel_envirD: "rel_envir p f g \<Longrightarrow> p (f x) (g x)"
by(auto simp add: rel_envir_def)

lemma rel_envirE [Pure.elim 2, elim]: "rel_envir p f g \<Longrightarrow> (p (f x) (g x) \<Longrightarrow> thesis) \<Longrightarrow> thesis"
by(simp add: rel_envir_conv)

lemma rel_envir_mono: "\<lbrakk> rel_envir p f g; \<And>x. p (f x) (g x) \<Longrightarrow> q (f' x) (g' x) \<rbrakk> \<Longrightarrow> rel_envir q f' g'"
by blast

lemma rel_envir_mono1: "\<lbrakk> pred_envir p f; \<And>x. p (f x) \<Longrightarrow> q (f' x) (g' x) \<rbrakk> \<Longrightarrow> rel_envir q f' g'"
by blast

lemma pred_envir_mono2: "\<lbrakk> rel_envir p f g; \<And>x. p (f x) (g x) \<Longrightarrow> q (f' x) \<rbrakk> \<Longrightarrow> pred_envir q f'"
by blast

end