theory GPV_Bisim imports
  Generative_Probabilistic_Value
begin

subsection {* Bisimulation for oracles *}

text \<open>Bisimulation is a consequence of parametricity\<close>

lemma exec_gpv_oracle_bisim':
  assumes *: "X s1 s2"
  and bisim: "\<And>s1 s2 x. X s1 s2 \<Longrightarrow> rel_spmf (\<lambda>(a, s1') (b, s2'). a = b \<and> X s1' s2') (oracle1 s1 x) (oracle2 s2 x)"
  shows "rel_spmf (\<lambda>(a, s1') (b, s2'). a = b \<and> X s1' s2') (exec_gpv oracle1 gpv s1) (exec_gpv oracle2 gpv s2)"
by(rule exec_gpv_parametric[of X "op =" "op =", unfolded gpv.rel_eq rel_prod_conv, THEN rel_funD, THEN rel_funD, THEN rel_funD, OF rel_funI refl, OF rel_funI *])(simp add: bisim)

lemma exec_gpv_oracle_bisim:
  assumes *: "X s1 s2"
  and bisim: "\<And>s1 s2 x. X s1 s2 \<Longrightarrow> rel_spmf (\<lambda>(a, s1') (b, s2'). a = b \<and> X s1' s2') (oracle1 s1 x) (oracle2 s2 x)"
  and R: "\<And>x s1' s2'. \<lbrakk> X s1' s2'; (x, s1') \<in> set_spmf (exec_gpv oracle1 gpv s1); (x, s2') \<in> set_spmf (exec_gpv oracle2 gpv s2) \<rbrakk> \<Longrightarrow> R (x, s1') (x, s2')"
  shows "rel_spmf R (exec_gpv oracle1 gpv s1) (exec_gpv oracle2 gpv s2)"
apply(rule spmf_rel_mono_strong)
apply(rule exec_gpv_oracle_bisim'[OF * bisim])
apply(auto dest: R)
done

lemma run_gpv_oracle_bisim:
  assumes  "X s1 s2"
  and "\<And>s1 s2 x. X s1 s2 \<Longrightarrow> rel_spmf (\<lambda>(a, s1') (b, s2'). a = b \<and> X s1' s2') (oracle1 s1 x) (oracle2 s2 x)"
  shows "run_gpv oracle1 gpv s1 = run_gpv oracle2 gpv s2"
using exec_gpv_oracle_bisim'[OF assms]
by(fold spmf_rel_eq)(fastforce simp add: spmf_rel_map intro: rel_spmf_mono)

lemma rel_spmf_bindI':
  "\<lbrakk> rel_spmf A p q; \<And>x y. \<lbrakk> A x y; x \<in> set_spmf p; y \<in> set_spmf q \<rbrakk> \<Longrightarrow> rel_spmf B (f x) (g y) \<rbrakk>
  \<Longrightarrow> rel_spmf B (p \<bind> f) (q \<bind> g)"
apply(rule rel_spmf_bindI[where R="\<lambda>x y. A x y \<and> x \<in> set_spmf p \<and> y \<in> set_spmf q"])
 apply(erule spmf_rel_mono_strong; simp)
apply simp
done

context
  fixes joint_oracle :: "('s1 \<times> 's2) \<Rightarrow> 'a \<Rightarrow> (('b \<times> 's1) \<times> ('b \<times> 's2)) spmf"
  and oracle1 :: "'s1 \<Rightarrow> 'a \<Rightarrow> ('b \<times> 's1) spmf"
  and bad1 :: "'s1 \<Rightarrow> bool"
  and oracle2 :: "'s2 \<Rightarrow> 'a \<Rightarrow> ('b \<times> 's2) spmf"
  and bad2 :: "'s2 \<Rightarrow> bool"
begin

partial_function (spmf) exec_until_bad :: "('x, 'a, 'b) gpv \<Rightarrow> 's1 \<Rightarrow> 's2 \<Rightarrow> (('x \<times> 's1) \<times> ('x \<times> 's2)) spmf"
where
  "exec_until_bad gpv s1 s2 =
   bind_spmf (the_gpv gpv) (\<lambda>generat.
     case generat of Pure x \<Rightarrow> return_spmf ((x, s1), (x, s2))
     | IO out f \<Rightarrow> bind_spmf (joint_oracle (s1, s2) out) (\<lambda>((x, s1'), (y, s2')).
       if bad1 s1' \<or> bad2 s2' then pair_spmf (exec_gpv oracle1 (f x) s1') (exec_gpv oracle2 (f y) s2')
       else exec_until_bad (f x) s1' s2'))"

lemma exec_until_bad_fixp_induct [case_names adm bottom step]:
  assumes "ccpo.admissible (fun_lub lub_spmf) (fun_ord (ord_spmf op =)) (\<lambda>f. P (\<lambda>gpv s1 s2. f ((gpv, s1), s2)))"
  and "P (\<lambda>_ _ _. return_pmf None)"
  and "\<And>exec_until_bad'. P exec_until_bad' \<Longrightarrow> 
     P (\<lambda>gpv s1 s2. bind_spmf (the_gpv gpv) (\<lambda>generat.
       case generat of Pure x \<Rightarrow> return_spmf ((x, s1), (x, s2))
       | IO out f \<Rightarrow> bind_spmf (joint_oracle (s1, s2) out) (\<lambda>((x, s1'), (y, s2')).
         if bad1 s1' \<or> bad2 s2' then pair_spmf (exec_gpv oracle1 (f x) s1') (exec_gpv oracle2 (f y) s2')
         else exec_until_bad' (f x) s1' s2')))"
  shows "P exec_until_bad"
using assms by(rule exec_until_bad.fixp_induct[unfolded curry_conv[abs_def]])

end


lemma exec_gpv_oracle_bisim_bad':
  fixes s1 :: 's1 and s2 :: 's2 and X :: "'s1 \<Rightarrow> 's2 \<Rightarrow> bool"
  and oracle1 :: "'s1 \<Rightarrow> 'a \<Rightarrow> ('b \<times> 's1) spmf"
  and oracle2 :: "'s2 \<Rightarrow> 'a \<Rightarrow> ('b \<times> 's2) spmf"
  assumes *: "X s1 s2"
  and bad: "bad1 s1 = bad2 s2"
  and bisim: "\<And>s1 s2 x. X s1 s2 \<Longrightarrow> rel_spmf (\<lambda>(a, s1') (b, s2'). bad1 s1' = bad2 s2' \<and> (\<not> bad2 s2' \<longrightarrow> a = b \<and> X s1' s2')) (oracle1 s1 x) (oracle2 s2 x)"
  and bad_sticky1: "callee_invariant oracle1 bad1"
  and bad_sticky2: "callee_invariant oracle2 bad2"
  and lossless1: "\<And>s1 x. bad1 s1 \<Longrightarrow> lossless_spmf (oracle1 s1 x)"
  and lossless2: "\<And>s2 x. bad2 s2 \<Longrightarrow> lossless_spmf (oracle2 s2 x)"
  and lossless: "lossless_gpv \<I> gpv"
  and WT_oracle1: "\<And>s1. \<I> \<turnstile>c oracle1 s1 \<surd>"
  and WT_oracle2: "\<And>s2. \<I> \<turnstile>c oracle2 s2 \<surd>"
  and WT_gpv: "\<I> \<turnstile>g gpv \<surd>"
  shows "rel_spmf (\<lambda>(a, s1') (b, s2'). bad1 s1' = bad2 s2' \<and> (\<not> bad2 s2' \<longrightarrow> a = b \<and> X s1' s2')) (exec_gpv oracle1 gpv s1) (exec_gpv oracle2 gpv s2)"
  (is "rel_spmf ?R ?p ?q")
proof -
  let ?R' = "\<lambda>(a, s1') (b, s2'). bad1 s1' = bad2 s2' \<and> (\<not> bad2 s2' \<longrightarrow> a = b \<and> X s1' s2')"
  from bisim have "\<forall>s1 s2 x. X s1 s2 \<longrightarrow> rel_spmf ?R' (oracle1 s1 x) (oracle2 s2 x)" by blast
  then obtain joint_oracle
    where oracle1 [symmetric]: "\<And>s1 s2 x. X s1 s2 \<Longrightarrow> map_spmf fst (joint_oracle s1 s2 x) = oracle1 s1 x"
    and oracle2 [symmetric]: "\<And>s1 s2 x. X s1 s2 \<Longrightarrow> map_spmf snd (joint_oracle s1 s2 x) = oracle2 s2 x"
    and 3 [rotated]: "\<And>s1 s2 x y y' s1' s2'. \<lbrakk> X s1 s2; ((y, s1'), (y', s2')) \<in> set_spmf (joint_oracle s1 s2 x) \<rbrakk>
      \<Longrightarrow> bad1 s1' = bad2 s2' \<and> (\<not> bad2 s2' \<longrightarrow> y = y' \<and> X s1' s2')"
    apply atomize_elim
    apply(unfold rel_spmf_simps all_conj_distrib[symmetric] all_simps(6) imp_conjR[symmetric])
    apply(subst choice_iff[symmetric] ex_simps(6))+
    apply fastforce
    done
  let ?joint_oracle = "\<lambda>(s1, s2). joint_oracle s1 s2"
  let ?pq = "exec_until_bad ?joint_oracle oracle1 bad1 oracle2 bad2 gpv s1 s2"

  have setD: "\<And>s1 s2 x y y' s1' s2'. \<lbrakk> X s1 s2; ((y, s1'), (y', s2')) \<in> set_spmf (joint_oracle s1 s2 x) \<rbrakk>
    \<Longrightarrow> (y, s1') \<in> set_spmf (oracle1 s1 x) \<and> (y', s2') \<in> set_spmf (oracle2 s2 x)"
    unfolding oracle1 oracle2 by(auto intro: rev_image_eqI)

  interpret oracle1: callee_invariant oracle1 bad1 by fact
  interpret oracle2: callee_invariant oracle2 bad2 by fact

  show ?thesis
  proof
    show "map_spmf fst ?pq = exec_gpv oracle1 gpv s1" using lossless WT_gpv *
    proof(induction arbitrary: s1 s2 rule: lossless_WT_gpv_induct)
      case (lossless_gpv p)
      have *: "\<And>out c s input s1 s2 input1 input2 s1' s2'. \<lbrakk> IO out c \<in> set_spmf p; 
                ((input1, s1'), (input2, s2')) \<in> set_spmf (joint_oracle s1 s2 out); X s1 s2; bad2 s \<rbrakk> 
              \<Longrightarrow> weight_spmf (exec_gpv oracle2 (c input2) s) = 1"
        using lossless_gpv.hyps(3) lossless2 lossless_gpv.hyps(4) unfolding lossless_spmf_def[symmetric]
        by(rule oracle2.lossless_exec_gpv)(auto simp add: WT_oracle2 dest!: setD WT_calleeD[OF WT_oracle2])
      show ?case using lossless_gpv.prems
        apply(rewrite exec_gpv.simps)
        apply(rewrite exec_until_bad.simps)
        apply(auto simp add: map_spmf_bind_spmf bind_map_spmf o_def oracle1 * 3 lossless_gpv.IH[OF _ WT_calleeD[OF WT_oracle1], OF _ setD[THEN conjunct1]] split: generat.split intro!: bind_spmf_cong[OF refl])
        done
    qed

    show "map_spmf snd ?pq = exec_gpv oracle2 gpv s2" using lossless WT_gpv *
    proof(induction arbitrary: s1 s2 rule: lossless_WT_gpv_induct)
      case (lossless_gpv p)
      have *: "\<And>out c s input s1 s2 input1 input2 s1' s2'. \<lbrakk> IO out c \<in> set_spmf p; 
                ((input1, s1'), (input2, s2')) \<in> set_spmf (joint_oracle s1 s2 out); X s1 s2; bad1 s \<rbrakk> 
              \<Longrightarrow> weight_spmf (exec_gpv oracle1 (c input1) s) = 1"
        using lossless_gpv.hyps(3) lossless1 lossless_gpv.hyps(4) unfolding lossless_spmf_def[symmetric]
        by(rule oracle1.lossless_exec_gpv)(auto simp add: WT_oracle1 dest!: setD WT_calleeD[OF WT_oracle1])
      show ?case using lossless_gpv.prems
        apply(rewrite exec_gpv.simps)
        apply(rewrite exec_until_bad.simps)
        apply(auto 4 3 simp add: map_spmf_bind_spmf bind_map_spmf o_def oracle2 * 3 lossless_gpv.IH[OF _ WT_calleeD[OF WT_oracle1], OF _ setD[THEN conjunct1]] split: generat.split intro!: bind_spmf_cong[OF refl] dest: 3)
        done
    qed

    have "set_spmf ?pq \<subseteq> {(as1, bs2). ?R' as1 bs2}" using * bad
    proof(induction arbitrary: gpv s1 s2 rule: exec_until_bad_fixp_induct)
      case adm show ?case by(intro cont_intro ccpo_class.admissible_leI)
      case bottom show ?case by simp
      case step
      show ?case using step.prems
        apply(clarsimp simp add: set_bind_spmf bind_UNION step.IH 3 del: subsetI intro!: UN_least split: generat.split)
        apply(auto dest!: oracle1.exec_gpv_invariant oracle2.exec_gpv_invariant 3)
        done
    qed
    then show "\<And>x y. (x, y) \<in> set_spmf ?pq \<Longrightarrow> ?R x y" by auto
  qed
qed

lemma exec_gpv_oracle_bisim_bad:
  assumes *: "X s1 s2"
  and bad: "bad1 s1 = bad2 s2"
  and bisim: "\<And>s1 s2 x. X s1 s2 \<Longrightarrow> rel_spmf (\<lambda>(a, s1') (b, s2'). bad1 s1' = bad2 s2' \<and> (\<not> bad2 s2' \<longrightarrow> a = b \<and> X s1' s2')) (oracle1 s1 x) (oracle2 s2 x)"
  and bad_sticky1: "callee_invariant oracle1 bad1"
  and bad_sticky2: "callee_invariant oracle2 bad2"
  and lossless1: "\<And>s1 x. bad1 s1 \<Longrightarrow> lossless_spmf (oracle1 s1 x)"
  and lossless2: "\<And>s2 x. bad2 s2 \<Longrightarrow> lossless_spmf (oracle2 s2 x)"
  and lossless: "lossless_gpv \<I> gpv"
  and WT_oracle1: "\<And>s1. \<I> \<turnstile>c oracle1 s1 \<surd>"
  and WT_oracle2: "\<And>s2. \<I> \<turnstile>c oracle2 s2 \<surd>"
  and WT_gpv: "\<I> \<turnstile>g gpv \<surd>"
  and R: "\<And>a s1 b s2. \<lbrakk> bad1 s1 = bad2 s2; \<not> bad2 s2 \<Longrightarrow> a = b \<and> X s1 s2 \<rbrakk> \<Longrightarrow> R (a, s1) (b, s2)"
  shows "rel_spmf R (exec_gpv oracle1 gpv s1) (exec_gpv oracle2 gpv s2)"
using exec_gpv_oracle_bisim_bad'[OF * bad bisim bad_sticky1 bad_sticky2 lossless1 lossless2 lossless WT_oracle1 WT_oracle2 WT_gpv]
by(rule rel_spmf_mono)(auto intro: R)

lemma exec_gpv_oracle_bisim_bad_top:
  assumes "X s1 s2"
  and "bad1 s1 = bad2 s2"
  and "\<And>s1 s2 x. X s1 s2 \<Longrightarrow> rel_spmf (\<lambda>(a, s1') (b, s2'). bad1 s1' = bad2 s2' \<and> (\<not> bad2 s2' \<longrightarrow> a = b \<and> X s1' s2')) (oracle1 s1 x) (oracle2 s2 x)"
  and "callee_invariant oracle1 bad1"
  and "callee_invariant oracle2 bad2"
  and "\<And>s1 x. bad1 s1 \<Longrightarrow> lossless_spmf (oracle1 s1 x)"
  and "\<And>s2 x. bad2 s2 \<Longrightarrow> lossless_spmf (oracle2 s2 x)"
  and "lossless_gpv \<I>_top gpv"
  and R: "\<And>a s1 b s2. \<lbrakk> bad1 s1 = bad2 s2; \<not> bad2 s2 \<Longrightarrow> a = b \<and> X s1 s2 \<rbrakk> \<Longrightarrow> R (a, s1) (b, s2)"
  shows "rel_spmf R (exec_gpv oracle1 gpv s1) (exec_gpv oracle2 gpv s2)"
using assms(1-8) _ _ _ R
by(rule exec_gpv_oracle_bisim_bad)(simp_all)

lemma identical_until_bad:
  assumes bad_eq: "map_spmf bad p = map_spmf bad q"
  and not_bad: "measure (measure_spmf (map_spmf (\<lambda>x. (f x, bad x)) p)) (A \<times> {False}) = measure (measure_spmf (map_spmf (\<lambda>x. (f x, bad x)) q)) (A \<times> {False})"
  shows "\<bar>measure (measure_spmf (map_spmf f p)) A - measure (measure_spmf (map_spmf f q)) A\<bar> \<le> spmf (map_spmf bad p) True"
proof -
  have "\<bar>ereal (measure (measure_spmf (map_spmf f p)) A) - ereal (measure (measure_spmf (map_spmf f q)) A)\<bar> = 
    \<bar>(\<integral>\<^sup>+ x. indicator A (f x) \<partial>measure_spmf p) - \<integral>\<^sup>+ x. indicator A (f x) \<partial>measure_spmf q\<bar>"
    unfolding measure_spmf.emeasure_eq_measure[symmetric]
    by(simp add: nn_integral_indicator[symmetric] indicator_vimage[abs_def] o_def)
  also have "\<dots> =
    \<bar>(\<integral>\<^sup>+ x. indicator (A \<times> {False}) (f x, bad x) + indicator (A \<times> {True}) (f x, bad x) \<partial>measure_spmf p) -
     (\<integral>\<^sup>+ x. indicator (A \<times> {False}) (f x, bad x) + indicator (A \<times> {True}) (f x, bad x) \<partial>measure_spmf q)\<bar>"
    by(intro arg_cong[where f=abs] arg_cong2[where f="op -"] nn_integral_cong)(simp_all split: split_indicator)
  also have "\<dots> = 
    \<bar>emeasure (measure_spmf (map_spmf (\<lambda>x. (f x, bad x)) p)) (A \<times> {False}) + (\<integral>\<^sup>+ x. indicator (A \<times> {True}) (f x, bad x) \<partial>measure_spmf p) -
     (emeasure (measure_spmf (map_spmf (\<lambda>x. (f x, bad x)) q)) (A \<times> {False}) + (\<integral>\<^sup>+ x. indicator (A \<times> {True}) (f x, bad x) \<partial>measure_spmf q))\<bar>"
    by(subst (1 2) nn_integral_add)(simp_all add: indicator_vimage[abs_def] o_def nn_integral_indicator[symmetric])
  also have "\<dots> = \<bar>(\<integral>\<^sup>+ x. indicator (A \<times> {True}) (f x, bad x) \<partial>measure_spmf p) - (\<integral>\<^sup>+ x. indicator (A \<times> {True}) (f x, bad x) \<partial>measure_spmf q)\<bar>"
    (is "_ = \<bar>?x - ?y\<bar>")
    by(simp add: measure_spmf.emeasure_eq_measure not_bad ereal_diff_add_eq_diff_diff_swap ereal_diff_add_assoc2 ereal_add_uminus_conv_diff)
  also have "\<dots> \<le> max ?x ?y"
  proof(rule ereal_abs_leI)
    have "?x - ?y \<le> ?x - 0" by(rule ereal_minus_mono)(simp_all add: nn_integral_nonneg)
    also have "\<dots> \<le> max ?x ?y" by simp
    finally show "?x - ?y \<le> \<dots>" .

    have "- (?x - ?y) = ?y - ?x"
      by(rule ereal_minus_diff_eq)(simp_all add: measure_spmf.nn_integral_indicator_neq_infty)
    also have "\<dots> \<le> ?y - 0" by(rule ereal_minus_mono)(simp_all add: nn_integral_nonneg)
    also have "\<dots> \<le> max ?x ?y" by simp
    finally show "- (?x - ?y) \<le> \<dots>" .
  qed
  also have "\<dots> \<le> max (\<integral>\<^sup>+ x. indicator {True} (bad x) \<partial>measure_spmf p) (\<integral>\<^sup>+ x. indicator {True} (bad x) \<partial>measure_spmf q)"
    by(intro ereal_max_mono nn_integral_mono)(simp_all split: split_indicator)
  also have "\<dots> = ereal (spmf (map_spmf bad p) True)"
    using arg_cong2[where f=spmf, OF bad_eq refl, of True, THEN arg_cong[where f=ereal]]
    unfolding ereal_spmf_map_conv_nn_integral indicator_vimage[abs_def] by simp
  finally show ?thesis by simp
qed

end
