theory Monad_Locale imports
  Misc
  "~~/src/HOL/Library/Monad_Syntax"
begin

context begin interpretation lifting_syntax .
lemma rel_fun_conversep:
  "(A^--1 ===> B^--1) = (A ===> B)^--1"
by(auto simp add: rel_fun_def fun_eq_iff)
end

lemma left_unique_Grp [iff]:
  "left_unique (BNF_Def.Grp A f) \<longleftrightarrow> inj_on f A"
unfolding Grp_def left_unique_def by(auto simp add: inj_on_def)

lemma right_unique_Grp [simp, intro!]: "right_unique (BNF_Def.Grp A f)"
by(simp add: Grp_def right_unique_def)

lemma bi_unique_Grp [iff]:
  "bi_unique (BNF_Def.Grp A f) \<longleftrightarrow> inj_on f A"
by(simp add: bi_unique_alt_def)

lemma left_total_Grp [iff]:
  "left_total (BNF_Def.Grp A f) \<longleftrightarrow> A = UNIV"
by(auto simp add: left_total_def Grp_def)

lemma right_total_Grp [iff]:
  "right_total (BNF_Def.Grp A f) \<longleftrightarrow> f ` A = UNIV"
by(auto simp add: right_total_def BNF_Def.Grp_def image_def)

lemma bi_total_Grp [iff]:
  "bi_total (BNF_Def.Grp A f) \<longleftrightarrow> A = UNIV \<and> surj f"
by(auto simp add: bi_total_alt_def)

lemma left_unique_vimage2p [simp]:
  "\<lbrakk> left_unique P; inj f \<rbrakk> \<Longrightarrow> left_unique (BNF_Def.vimage2p f g P)"
unfolding vimage2p_Grp by(intro left_unique_OO) simp_all

lemma right_unique_vimage2p [simp]:
  "\<lbrakk> right_unique P; inj g \<rbrakk> \<Longrightarrow> right_unique (BNF_Def.vimage2p f g P)"
unfolding vimage2p_Grp by(intro right_unique_OO) simp_all

lemma bi_unique_vimage2p [simp]:
  "\<lbrakk> bi_unique P; inj f; inj g \<rbrakk> \<Longrightarrow> bi_unique (BNF_Def.vimage2p f g P)"
unfolding bi_unique_alt_def by simp

lemma left_total_vimage2p [simp]:
  "\<lbrakk> left_total P; surj g \<rbrakk> \<Longrightarrow> left_total (BNF_Def.vimage2p f g P)"
unfolding vimage2p_Grp by(intro left_total_OO) simp_all

lemma right_total_vimage2p [simp]:
  "\<lbrakk> right_total P; surj f \<rbrakk> \<Longrightarrow> right_total (BNF_Def.vimage2p f g P)"
unfolding vimage2p_Grp by(intro right_total_OO) simp_all

lemma bi_total_vimage2p [simp]:
  "\<lbrakk> bi_total P; surj f; surj g \<rbrakk> \<Longrightarrow> bi_total (BNF_Def.vimage2p f g P)"
unfolding bi_total_alt_def by simp

lemma vimage2p_eq [simp]:
  "inj f \<Longrightarrow> BNF_Def.vimage2p f f op = = op ="
by(auto simp add: vimage2p_def fun_eq_iff inj_on_def)

lemma vimage2p_conversep: "BNF_Def.vimage2p f g R^--1 = (BNF_Def.vimage2p g f R)^--1"
by(simp add: vimage2p_def fun_eq_iff)

named_theorems record_parametricity "unfold theorems for record_parametricity"
named_theorems record_typedef "type definition theorems for record parametricity"

context begin interpretation lifting_syntax .
lemma vimage2p_abs_rep:
  assumes h: "(Y ===> X1 ===> X2) h1 h2"
  and 1: "type_definition f1 g1 UNIV"
  and 2: "type_definition f2 g2 UNIV"
  shows "(Y ===> BNF_Def.vimage2p f1 f2 X1 ===> BNF_Def.vimage2p f1 f2 X2) (\<lambda>x. g1 \<circ> (h1 x \<circ> f1)) (\<lambda>x. g2 \<circ> (h2 x \<circ> f2))"
apply(rule rel_funI)
apply(rule comp_transfer[THEN rel_funD, THEN rel_funD])
 apply(rule Abs_transfer[OF 1 2])
apply(rule comp_transfer[THEN rel_funD, THEN rel_funD])
 apply(erule rel_funD[OF h])
apply(rule vimage2p_rel_fun)
done


lemma apsnd_parametric [transfer_rule]:
  "((A ===> B) ===> rel_prod C A ===> rel_prod C B) apsnd apsnd"
unfolding apsnd_def[abs_def] by transfer_prover

end
method record_parametricity uses unfold typedefs =
  (unfold unfold record_parametricity
      iso_tuple_fst_def iso_tuple_snd_def tuple_iso_tuple_def
      iso_tuple_fst_update_def[abs_def] iso_tuple_snd_update_def[abs_def]
      iso_tuple_cons_def
      repr.simps abst.simps o_assoc[symmetric] id_o o_id,
   ((assumption | rule
     typedefs record_typedef
     fst_transfer[THEN comp_transfer[THEN rel_funD, THEN rel_funD]]
     snd_transfer[THEN comp_transfer[THEN rel_funD, THEN rel_funD]]
     comp_transfer[THEN rel_funD, THEN rel_funD]
     apfst_parametric
     apsnd_parametric
     curry_transfer[THEN rel_funD, THEN rel_funD]
     curry_transfer[THEN rel_funD, THEN rel_funD, THEN rel_funD, OF id_transfer]
     vimage2p_abs_rep
     Abs_transfer
     vimage2p_rel_fun
     rel_funI
    )+)?
  )

attribute_setup locale_witness = \<open>Scan.succeed Locale.witness_add\<close>

section \<open>Monomorphic monads and monad transformers\<close>

subsection \<open>Plain monads\<close>

record ('a, 'm) monad =
  bind :: "'m \<Rightarrow> ('a \<Rightarrow> 'm) \<Rightarrow> 'm"
  return :: "'a \<Rightarrow> 'm"

context begin local_setup \<open>Local_Theory.map_background_naming (Name_Space.mandatory_path "monad")\<close>
definition lift :: "('a, 'm, 'more) monad_ext \<Rightarrow> ('a \<Rightarrow> 'a) \<Rightarrow> 'm \<Rightarrow> 'm"
where "lift m f x = bind m x (return m \<circ> f)"
end

locale monad_syntax = fixes m :: "('a, 'm, 'more) monad_scheme" begin
adhoc_overloading Monad_Syntax.bind "bind m"
end

locale monad = fixes m :: "('a, 'm, 'more) monad_scheme"
  assumes bind_assoc: "\<And>(x :: 'm) f g. bind m (bind m x f) g = bind m x (\<lambda>y. bind m (f y) g)" 
  and return_bind: "\<And>x f. bind m (return m x) f = f x"
  and bind_return: "\<And>x. bind m x (return m) = x"

lemma inj_Rep_monad_ext [simp]: "inj Rep_monad_ext"
by(simp add: inj_on_def Rep_monad_ext_inject)

lemma surj_Rep_monad_ext [simp]: "surj Rep_monad_ext"
by(rule surjI[where f=Abs_monad_ext])(simp add: Abs_monad_ext_inverse)


context begin interpretation lifting_syntax .

definition rel_monad_ext
  :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('m1 \<Rightarrow> 'm2 \<Rightarrow> bool) \<Rightarrow> ('more1 \<Rightarrow> 'more2 \<Rightarrow> bool)
  \<Rightarrow> ('a, 'm1, 'more1) monad_ext \<Rightarrow> ('b, 'm2, 'more2) monad_ext \<Rightarrow> bool"
where [record_parametricity]:
  "rel_monad_ext A M X = 
  BNF_Def.vimage2p Rep_monad_ext Rep_monad_ext (rel_prod (rel_prod (M ===> (A ===> M) ===> M) (A ===> M)) X)"

declare
  monad_ext_Tuple_Iso_def [record_parametricity]
  monad.select_defs [record_parametricity]
  monad.update_defs [record_parametricity]
  type_definition_monad_ext [record_typedef]

abbreviation rel_monad :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('m1 \<Rightarrow> 'm2 \<Rightarrow> bool)
  \<Rightarrow> ('a, 'm1) monad \<Rightarrow> ('b, 'm2) monad \<Rightarrow> bool"
where "rel_monad A M == rel_monad_ext A M op ="

lemma rel_monad_ext_conversep:
  "rel_monad_ext A^--1 M^--1 X^--1 = (rel_monad_ext A M X)^--1"
unfolding rel_monad_ext_def rel_fun_conversep prod.rel_conversep vimage2p_conversep ..

lemma left_unique_rel_monad_ext [transfer_rule]:
  "\<lbrakk> left_unique X; left_unique A; left_total A; left_unique M; left_total M \<rbrakk>
  \<Longrightarrow> left_unique (rel_monad_ext A M X)"
unfolding rel_monad_ext_def
by(simp add: prod.left_unique_rel left_unique_fun left_total_fun)

lemma right_unique_rel_monad_ext [transfer_rule]:
  "\<lbrakk> right_unique X; right_unique A; right_total A; right_unique M; right_total M \<rbrakk>
  \<Longrightarrow> right_unique (rel_monad_ext A M X)"
unfolding rel_monad_ext_def
by(simp add: prod.right_unique_rel right_unique_fun right_total_fun)

lemma left_total_rel_monad_ext [transfer_rule]:
  "\<lbrakk> left_total X; left_unique A; left_total A; left_unique M; left_total M \<rbrakk>
  \<Longrightarrow> left_total (rel_monad_ext A M X)"
unfolding rel_monad_ext_def
by(simp add: prod.left_total_rel left_total_fun left_unique_fun)

lemma right_total_rel_monad_ext [transfer_rule]:
  "\<lbrakk> right_total X; right_unique A; right_total A; right_unique M; right_total M \<rbrakk>
  \<Longrightarrow> right_total (rel_monad_ext A M X)"
unfolding rel_monad_ext_def
by(simp add: prod.right_total_rel right_total_fun right_unique_fun)

lemma rel_monad_ext_eq [relator_eq]:
  "rel_monad_ext op = op = op = = op ="
unfolding rel_monad_ext_def rel_fun_eq prod.rel_eq
by(simp add: eq_OO eq_alt[symmetric])

lemma bind_monad_parametric [transfer_rule]:
  "(rel_monad_ext A M X ===> M ===> (A ===> M) ===> M) bind bind"
by(record_parametricity)

lemma return_monad_parametric [transfer_rule]:
  "(rel_monad_ext A M X ===> A ===> M) return return"
by(record_parametricity)

lemma more_monad_parametric [transfer_rule]:
  "(rel_monad_ext A M X ===> X) more more"
by(record_parametricity)

lemma lift_monad_parametric [transfer_rule]:
  "(rel_monad_ext A M X ===> (A ===> A) ===> M ===> M) monad.lift monad.lift"
unfolding monad.lift_def[abs_def] by transfer_prover

lemma bind_update_parametric [transfer_rule]:
  "(((M ===> (A ===> M) ===> M) ===> (M ===> (A ===> M) ===> M))
  ===> rel_monad_ext A M X ===> rel_monad_ext A M X) bind_update bind_update"
by record_parametricity

lemma return_update_parametric [transfer_rule]:
  "(((A ===> M) ===> (A ===> M)) ===> rel_monad_ext A M X ===> rel_monad_ext A M X)
   return_update return_update"
by record_parametricity

lemma more_update_monad_parametric [transfer_rule]:
  "((X ===> X') ===> rel_monad_ext A M X ===> rel_monad_ext A M X')
   more_update more_update"
by record_parametricity

lemma monad_ext_parametric [transfer_rule]:
  "((M ===> (A ===> M) ===> M) ===> (A ===> M) ===> X ===> rel_monad_ext A M X)
   monad_ext monad_ext"
by(record_parametricity unfold: monad_ext_def [abs_def])

lemma make_monad_parametric [transfer_rule]:
  "((M ===> (A ===> M) ===> M) ===> (A ===> M) ===> rel_monad A M) make make"
unfolding monad.defs[abs_def] by transfer_prover

lemma extend_monad_parametric [transfer_rule]:
  "(rel_monad A M ===> X ===> rel_monad_ext A M X) extend extend"
unfolding monad.defs[abs_def] by transfer_prover

lemma fields_monad_parametric [transfer_rule]:
  "((M ===> (A ===> M) ===> M) ===> (A ===> M) ===> rel_monad A M) fields fields"
unfolding monad.defs[abs_def] by transfer_prover

lemma truncate_monad_parametric [transfer_rule]:
  "(rel_monad_ext A M X ===> rel_monad A M) truncate truncate"
unfolding monad.defs[abs_def] by transfer_prover

end


subsubsection \<open>Identity monad\<close>

definition identity_monad_ext :: "'more \<Rightarrow> ('a, 'a, 'more) monad_ext"
where "\<And>more. identity_monad_ext more = \<lparr>bind = \<lambda>x f. f x, return = \<lambda>x. x, \<dots> = more\<rparr>"

abbreviation identity_monad :: "('a, 'a) monad"
where "identity_monad \<equiv> identity_monad_ext ()"

lemma identity_sel [simp]:
  shows bind_identity: "\<And>more. bind (identity_monad_ext more) = (\<lambda>x f. f x)"
  and return_identity: "\<And>more. return (identity_monad_ext more) = (\<lambda>x. x)"
  and more_identity: "\<And>more. monad.more (identity_monad_ext more) = more"
by(simp_all add: identity_monad_ext_def)

lemma monad_identity [simp, intro!, locale_witness]: "\<And>more. monad (identity_monad_ext more)"
by unfold_locales simp_all

context begin interpretation lifting_syntax .
lemma identity_monad_parametric [transfer_rule]:
  "(X ===> rel_monad_ext A A X) identity_monad_ext identity_monad_ext"
unfolding identity_monad_ext_def[abs_def] by transfer_prover
end

subsection \<open>Monads with failure\<close>

record ('a, 'm) monad_fail = "('a, 'm) monad" +
  fail :: "'m"

copy_bnf ('m, 'mpre) monad_fail_ext

declare
  type_definition_monad_fail_ext [record_typedef]
  rel_monad_fail_ext_def [record_parametricity]
  monad_fail_ext_Tuple_Iso_def [record_parametricity]
  monad_fail.select_defs [record_parametricity]
  monad_fail.update_defs [record_parametricity]

abbreviation rel_monad_fail_scheme
  :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('m1 \<Rightarrow> 'm2 \<Rightarrow> bool) \<Rightarrow> ('more1 \<Rightarrow> 'more2 \<Rightarrow> bool)
  \<Rightarrow> ('a, 'm1, 'more1) monad_fail_scheme \<Rightarrow> ('b, 'm2, 'more2) monad_fail_scheme \<Rightarrow> bool"
where "rel_monad_fail_scheme A M X \<equiv> rel_monad_ext A M (rel_monad_fail_ext M X)"

abbreviation rel_monad_fail
  :: "('a \<Rightarrow> 'b \<Rightarrow> bool) \<Rightarrow> ('m1 \<Rightarrow> 'm2 \<Rightarrow> bool)
  \<Rightarrow> ('a, 'm1) monad_fail \<Rightarrow> ('b, 'm2) monad_fail \<Rightarrow> bool"
where "rel_monad_fail A M \<equiv> rel_monad_fail_scheme A M op ="

context begin interpretation lifting_syntax .
lemma fail_monad_parametric [transfer_rule]:
  "(rel_monad_fail_scheme A M X ===> M) fail fail"
by record_parametricity

lemma more_monad_fail_parametric [transfer_rule]:
  "(rel_monad_fail_scheme A M X ===> X) more more"
by record_parametricity

lemma fail_update_parametric [transfer_rule]:
  "((M ===> M) ===> rel_monad_fail_scheme A M X ===> rel_monad_fail_scheme A M X)
   fail_update fail_update"
by record_parametricity

lemma more_update_monad_fail_parametric [transfer_rule]:
  "((X ===> X') ===> rel_monad_fail_scheme A M X ===> rel_monad_fail_scheme A M X') more_update more_update"
by record_parametricity

lemma monad_fail_ext_parametric [transfer_rule]:
  "(M ===> X ===> rel_monad_fail_ext M X) monad_fail_ext monad_fail_ext"
by(record_parametricity unfold: monad_fail_ext_def [abs_def])

lemma make_monad_fail_parametric [transfer_rule]:
  "((M ===> (A ===> M) ===> M) ===> (A ===> M) ===> M ===> rel_monad_fail A M) make make"
unfolding monad_fail.defs[abs_def] by transfer_prover

lemma extend_monad_fail_parametric [transfer_rule]:
  "(rel_monad_fail A M ===> X ===> rel_monad_fail_scheme A M X) extend extend"
unfolding monad_fail.defs[abs_def] by transfer_prover

lemma fields_monad_fail_parametric [transfer_rule]:
  "(M ===> rel_monad_fail_ext M op =) fields fields"
unfolding monad_fail.defs[abs_def] by transfer_prover

lemma truncate_monad_fail_parametric [transfer_rule]:
  "(rel_monad_fail_scheme A M X ===> rel_monad_fail A M) truncate truncate"
unfolding monad_fail.defs[abs_def] by transfer_prover

end

context begin local_setup \<open>Local_Theory.map_background_naming (Name_Space.mandatory_path "monad")\<close>
definition bind_option :: "('a, 'm, 'more) monad_fail_scheme \<Rightarrow> 'b option \<Rightarrow> ('b \<Rightarrow> 'm) \<Rightarrow> 'm"
where "bind_option m x f = (case x of None \<Rightarrow> fail m | Some x' \<Rightarrow> f x')"

simps_of_case bind_option_simps [simp]: bind_option_def

interpretation lifting_syntax .
lemma bind_option_parametric [transfer_rule]:
  "(rel_monad_fail_scheme A M X ===> rel_option B ===> (B ===> M) ===> M) bind_option bind_option"
unfolding bind_option_def[abs_def] by transfer_prover
end

locale monad_fail_syntax = monad_syntax m for m :: "('a, 'm, 'more) monad_fail_scheme" begin
adhoc_overloading Monad_Syntax.bind "monad.bind_option m"
end

locale monad_fail = monad m for m :: "('a, 'm, 'more) monad_fail_scheme" +
  assumes fail_bind: "\<And>f. bind m (fail m) f = fail m"


subsection \<open> The option monad \<close>

definition option_monad_ext :: "'more \<Rightarrow> ('a, 'a option, 'more) monad_fail_scheme"
where "\<And>more. option_monad_ext more = \<lparr>bind = Option.bind, return = Some, fail = None, \<dots> = more\<rparr>"

abbreviation option_monad :: "('a, 'a option) monad_fail" where "option_monad \<equiv> option_monad_ext ()"

lemma option_monad_sel [simp]:
  shows bind_option_monad: "\<And>more. bind (option_monad_ext more) = Option.bind"
  and return_option_monad: "\<And>more. return (option_monad_ext more) = Some"
  and fail_option_monad: "\<And>more. fail (option_monad_ext more) = None"
  and more_option_monad: "\<And>more. monad_fail.more (option_monad_ext more) = more"
by(simp_all add: option_monad_ext_def)

lemma option_monad [simp, intro!, locale_witness]: "\<And>more. monad (option_monad_ext more)"
by unfold_locales simp_all

lemma option_monad_fail [simp, intro!, locale_witness]: "\<And>more. monad_fail (option_monad_ext more)"
by unfold_locales simp

context begin interpretation lifting_syntax .
lemma option_monad_ext_parametric [transfer_rule]:
  "(X ===> rel_monad_fail_scheme A (rel_option A) X) option_monad_ext option_monad_ext"
unfolding option_monad_ext_def[abs_def] by transfer_prover
end

lemma bind_option_option [simp]: "\<And>more. monad.bind_option (option_monad_ext more) = Option.bind"
by(simp add: monad.bind_option_def fun_eq_iff split: option.split)

lemma monad_lift_option [simp]: "\<And>more. monad.lift option_monad = map_option"
by(simp add: monad.lift_def map_conv_bind_option fun_eq_iff)

subsection \<open> The set monad \<close>

definition set_monad_ext :: "'more \<Rightarrow> ('a, 'a set, 'more) monad_fail_scheme"
where "\<And>more. set_monad_ext more = \<lparr>bind = UNION, return = \<lambda>x. {x}, fail = {}, \<dots> = more\<rparr>"

abbreviation set_monad :: "('a, 'a set) monad_fail" where "set_monad \<equiv> set_monad_ext ()"

lemma set_monad_sel [simp]:
  shows bind_set_monad: "\<And>more. bind (set_monad_ext more) = UNION"
  and return_set_monad: "\<And>more. return (set_monad_ext more) = (\<lambda>x. {x})"
  and fail_set_monad: "\<And>more. fail (set_monad_ext more) = {}"
  and more_set_monad: "\<And>more. monad_fail.more (set_monad_ext more) = more"
by(simp_all add: set_monad_ext_def)

lemma set_monad [simp, intro!, locale_witness]: "\<And>more. monad (set_monad_ext more)"
by unfold_locales auto

lemma set_monad_fail [simp, intro!, locale_witness]: "\<And>more. monad_fail (set_monad_ext more)"
by unfold_locales simp

context begin interpretation lifting_syntax .
lemma set_monad_ext_parametric [transfer_rule]:
  "(X ===> rel_monad_fail_scheme A (rel_set A) X) set_monad_ext set_monad_ext"
unfolding set_monad_ext_def[abs_def] by transfer_prover
end

lemma bind_option_set [simp]: "\<And>more. monad.bind_option (set_monad_ext more) = (\<lambda>x. UNION (set_option x))"
by(simp add: monad.bind_option_def fun_eq_iff split: option.split)

lemma monad_lift_set [simp]: "\<And>more. monad.lift (set_monad_ext more) = image"
by(simp add: monad.lift_def image_eq_UN o_def fun_eq_iff)


subsection \<open> Monad homomorphism\<close>

locale monad_hom = m1: monad m1 + m2: monad m2
  for m1 :: "('a, 'm1, 'more1) monad_scheme"
  and m2 :: "('a, 'm2, 'more2) monad_scheme"
  and h :: "'m1 \<Rightarrow> 'm2"
  +
  assumes hom_return: "\<And>x. h (return m1 x) = return m2 x"
  and hom_bind: "\<And>x f. h (bind m1 x f) = bind m2 (h x) (h \<circ> f)"
begin

lemma hom_lift: "h (monad.lift m1 f x) = monad.lift m2 f (h x)"
by(simp add: monad.lift_def hom_bind hom_return o_def)

end

locale monad_fail_hom = m1: monad_fail m1 + m2: monad_fail m2 + monad_hom m1 m2 h
  for m1 :: "('a, 'm1, 'more1) monad_fail_scheme"
  and m2 :: "('a, 'm2, 'more2) monad_fail_scheme"
  and h :: "'m1 \<Rightarrow> 'm2"
  +
  assumes hom_fail: "h (fail m1) = fail m2"
begin

lemma hom_bind_option: "h (monad.bind_option m1 x f) = monad.bind_option m2 x (h \<circ> f)"
by(cases x)(simp_all add: hom_fail)

end

section \<open> Monad transformers \<close>

subsection \<open>The state monad transformer\<close>

datatype ('s, 'm) ST = ST (run_state: "'s \<Rightarrow> 'm")
  for rel: rel_ST'

hide_const (open) ST

primrec bind_state :: "('a \<times> 's, 'm, 'more) monad_scheme \<Rightarrow> ('s, 'm) ST \<Rightarrow> ('a \<Rightarrow> ('s, 'm) ST) \<Rightarrow> ('s, 'm) ST"
where "bind_state m (ST.ST x) f = ST.ST (\<lambda>s. bind m (x s) (\<lambda>(a, s'). run_state (f a) s'))"

lemma run_bind_state [simp]: "run_state (bind_state m x f) s = bind m (run_state x s) (\<lambda>(a, s'). run_state (f a) s')"
by(cases x) simp

definition return_state :: "('a \<times> 's, 'm, 'more) monad_scheme \<Rightarrow> 'a \<Rightarrow> ('s, 'm) ST"
where "return_state m x = ST.ST (\<lambda>s. return m (x, s))"

lemma run_return_state [simp]: "run_state (return_state m x) s = return m (x, s)"
by(simp add: return_state_def)

context
  fixes m :: "('a \<times> 's, 'm, 'more) monad_scheme"
  assumes "monad m"
begin
interpretation monad m by fact

lemma bind_state_assoc [simp]: "bind_state m (bind_state m x f) g = bind_state m x (\<lambda>y. bind_state m (f y) g)"
by(rule ST.expand ext)+(simp add: bind_assoc split_def)

lemma return_bind_state [simp]: "bind_state m (return_state m x) f = f x"
by(rule ST.expand ext)+(simp add: return_bind)

lemma bind_return_state [simp]: "bind_state m x (return_state m) = x"
by(rule ST.expand ext)+(simp add: bind_return)

end

context fixes m :: "('a \<times> 's, 'm, 'more) monad_fail_scheme" begin

definition fail_state :: "('s, 'm) ST"
where "fail_state = ST.ST (\<lambda>_. fail m)"

lemma run_fail_state [simp]: "run_state fail_state s = fail m"
by(simp add: fail_state_def)

context assumes "monad_fail m" begin
interpretation monad_fail m by fact

lemma fail_bind_state [simp]: "bind_state m fail_state f = fail_state"
by(rule ST.expand ext)+(simp add: fail_bind)

end
end

context begin interpretation lifting_syntax .

inductive rel_ST :: "('s1 \<Rightarrow> 's2 \<Rightarrow> bool) \<Rightarrow> ('m1 \<Rightarrow> 'm2 \<Rightarrow> bool) \<Rightarrow> ('s1, 'm1) ST \<Rightarrow> ('s2, 'm2) ST \<Rightarrow> bool"
  for S M x y
where "(S ===> M) (run_state x) (run_state y) \<Longrightarrow> rel_ST S M x y"

inductive_simps rel_ST_ST1: "rel_ST S M (ST.ST x) y"
inductive_simps rel_ST_ST2: "rel_ST S M x (ST.ST y)"
inductive_simps rel_ST_simps [simp]: "rel_ST S M (ST.ST x) (ST.ST y)"

lemma rel_ST_alt_def: "rel_ST S M = BNF_Def.vimage2p run_state run_state (S ===> M)"
apply(rule ext)+
subgoal for x y
  apply(cases x; cases y)
  apply(simp add: rel_ST.simps vimage2p_def)
  done
done

lemma inj_run_state [simp]: "inj run_state"
by(auto simp add: inj_on_def intro: ST.expand)

lemma surj_run_state [simp]: "surj run_state"
by(rule surjI[where f="ST.ST"]) simp

lemma rel_ST_eq [relator_eq]: "rel_ST op = op = = op ="
unfolding rel_ST_alt_def rel_fun_eq by (simp add: eq_OO eq_alt[symmetric])

lemma left_unique_rel_ST [transfer_rule]:
  "\<lbrakk> left_total S; left_unique M \<rbrakk> \<Longrightarrow> left_unique (rel_ST S M)"
unfolding rel_ST_alt_def by(simp add: left_unique_fun)

lemma right_unique_rel_ST [transfer_rule]:
  "\<lbrakk> right_total S; right_unique M \<rbrakk> \<Longrightarrow> right_unique (rel_ST S M)"
unfolding rel_ST_alt_def by(simp add: right_unique_fun)

lemma bi_unique_rel_ST [transfer_rule]:
  "\<lbrakk> bi_total S; bi_unique M \<rbrakk> \<Longrightarrow> bi_unique (rel_ST S M)"
unfolding bi_unique_alt_def bi_total_alt_def by(simp add: left_unique_rel_ST right_unique_rel_ST)

lemma right_total_rel_ST [transfer_rule]:
  "\<lbrakk> right_unique S; right_total M \<rbrakk> \<Longrightarrow> right_total (rel_ST S M)"
unfolding rel_ST_alt_def by(simp add: right_total_fun)

lemma left_total_rel_ST [transfer_rule]:
  "\<lbrakk> left_unique S; left_total M \<rbrakk> \<Longrightarrow> left_total (rel_ST S M)"
unfolding rel_ST_alt_def by(simp add: left_total_fun)

lemma bi_total_rel_ST [transfer_rule]:
  "\<lbrakk> bi_unique S; bi_total M \<rbrakk> \<Longrightarrow> bi_total (rel_ST S M)"
unfolding bi_unique_alt_def bi_total_alt_def by(simp add: left_total_rel_ST right_total_rel_ST)

lemma run_state_parametric [transfer_rule]: "(rel_ST S M ===> S ===> M) run_state run_state"
by(rule rel_funI)(erule rel_ST.cases)

context begin local_setup \<open>Local_Theory.map_background_naming (Name_Space.mandatory_path "ST")\<close>
lemma ST_parametric [transfer_rule]: "((S ===> M) ===> rel_ST S M) ST.ST ST.ST"
by(rule rel_funI)(rule rel_ST.intros; simp)
end

lemma rec_ST_parametric [transfer_rule]:
  "(((S ===> M) ===> C) ===> rel_ST S M ===> C) rec_ST rec_ST"
apply(rule rel_funI)+
subgoal for f g x y
  apply(cases x)
  apply(cases y)
  apply(auto dest: rel_funD)
  done
done

lemma bind_state_parametric [transfer_rule]:
  "(rel_monad_ext (rel_prod A S) M X ===> rel_ST S M ===> (A ===> rel_ST S M) ===> rel_ST S M) bind_state bind_state"
unfolding bind_state_def[abs_def] by transfer_prover

lemma return_state_parametric [transfer_rule]:
  "(rel_monad_ext (rel_prod A S) M X ===> A ===> rel_ST S M) return_state return_state"
unfolding return_state_def[abs_def] by transfer_prover

lemma fail_state_parametric [transfer_rule]:
  "(rel_monad_fail_scheme (rel_prod A S) M X ===> rel_ST S M) fail_state fail_state"
unfolding fail_state_def[abs_def] by transfer_prover

end

definition ST_ext :: "(('a \<times> 's, 'm, 'more) monad_scheme \<Rightarrow> 'more') 
  \<Rightarrow> ('a \<times> 's, 'm, 'more) monad_scheme \<Rightarrow> ('a, ('s, 'm) ST, 'more') monad_scheme"
where
  "ST_ext MORE m = \<lparr>bind = bind_state m, return = return_state m, \<dots> = MORE m\<rparr>"

abbreviation ST_scheme  :: "(('a \<times> 's, 'm, 'more) monad_scheme \<Rightarrow> 'more') 
  \<Rightarrow> ('a \<times> 's, 'm, 'more) monad_scheme \<Rightarrow> ('a, ('s, 'm) ST, 'more') monad_scheme"
where "ST_scheme \<equiv> ST_ext"

abbreviation ST :: "('a \<times> 's, 'm, 'more) monad_scheme \<Rightarrow> ('a, ('s, 'm) ST, 'more) monad_scheme"
where "ST \<equiv> ST_scheme monad.more"

lemma ST_sel:
  shows bind_ST: "bind (ST_scheme MORE m) = bind_state m"
  and return_ST: "return (ST_scheme MORE m) = return_state m"
  and more_ST [simp]: "monad.more (ST_scheme MORE m) = MORE m"
by(simp_all add: ST_ext_def)

lemma monad_lift_ST [simp]:
  "\<And>more. monad.lift (ST_scheme more m) f (ST.ST x) = ST.ST (monad.lift m (apfst f) \<circ> x)"
by(simp add: monad.lift_def ST_sel o_def split_def apfst_def map_prod_def)

lemma run_bind_ST [simp]:
  "\<And>more. run_state (bind (ST_scheme more m) x f) s = 
  bind m (run_state x s) (\<lambda>(a, s'). run_state (f a) s')"
by(simp add: bind_ST)

lemma run_return_ST [simp]: "\<And>more. run_state (return (ST_scheme more m) x) s = return m (x, s)"
by(simp add: return_ST)

lemma run_lift_ST [simp]:
  "\<And>more. run_state (monad.lift (ST_scheme more m) f x) s = monad.lift m (apfst f) (run_state x s)"
by(cases x) simp

lemma monad_ST [simp, intro!, locale_witness]: "\<And>more. monad m \<Longrightarrow> monad (ST_scheme more m)"
by unfold_locales(simp_all add: ST_sel)

context begin interpretation lifting_syntax .
lemma ST_ext_parametric [transfer_rule]:
  "((rel_monad_ext (rel_prod A S) M X ===> X') ===> rel_monad_ext (rel_prod A S) M X 
  ===> rel_monad_ext A (rel_ST S M) X')
    ST_ext ST_ext"
unfolding ST_ext_def[abs_def] by transfer_prover
end

definition ST_fail_ext :: "(('a \<times> 's, 'm, 'more) monad_fail_scheme \<Rightarrow> 'more')
  \<Rightarrow> ('a \<times> 's, 'm, 'more) monad_fail_scheme \<Rightarrow> (('s, 'm) ST, 'more') monad_fail_ext"
where "ST_fail_ext MORE m = \<lparr>fail = fail_state m, \<dots> = MORE m\<rparr>"

abbreviation ST_fail_scheme :: "(('a \<times> 's, 'm, 'more) monad_fail_scheme \<Rightarrow> 'more')
  \<Rightarrow> ('a \<times> 's, 'm, 'more) monad_fail_scheme \<Rightarrow> ('a, ('s, 'm) ST, 'more') monad_fail_scheme"
where "ST_fail_scheme MORE \<equiv> ST_scheme (ST_fail_ext MORE)"

abbreviation ST_fail :: "('a \<times> 's, 'm, 'more) monad_fail_scheme \<Rightarrow> ('a, ('s, 'm) ST, 'more) monad_fail_scheme"
where "ST_fail \<equiv> ST_fail_scheme monad_fail.more"

lemma ST_fail_sel:
  shows fail_ST_fail: "\<And>more. fail (ST_fail_scheme more m) = fail_state m"
  and more_ST_fail [simp]: "\<And>more. monad_fail.more (ST_fail_scheme more m) = more m"
by(simp_all add: ST_fail_ext_def ST_ext_def)

lemma run_fail_ST [simp]: "\<And>more. run_state (fail (ST_fail_scheme more m)) s = fail m"
by(simp add: fail_ST_fail)

lemma monad_fail_ST [simp, intro!, locale_witness]:
  fixes more m
  assumes "monad_fail m"
  shows "monad_fail (ST_fail_scheme more m)"
proof -
  interpret monad_fail m by fact
  show ?thesis by unfold_locales(rule ST.expand, simp add: fail_bind fun_eq_iff)
qed

lemma run_bind_option_ST [simp]:
  "\<And>more. run_state (monad.bind_option (ST_fail_scheme more m) x f) s = 
  monad.bind_option m x (\<lambda>y. run_state (f y) s)"
by(cases x) simp_all


subsubsection \<open> Transformation of homomorphisms \<close>

primrec hom_state :: "('m1 \<Rightarrow> 'm2) \<Rightarrow> ('s, 'm1) ST \<Rightarrow> ('s, 'm2) ST"
where "hom_state h (ST.ST x) = ST.ST (\<lambda>s. h (x s))"

lemma run_hom_state [simp]: "run_state (hom_state h x) s = h (run_state x s)"
by(cases x)(simp)

context begin interpretation lifting_syntax .
lemma hom_state_parametric[transfer_rule]:
  "((M1 ===> M2) ===> rel_ST S M1 ===> rel_ST S M2) hom_state hom_state"
unfolding hom_state_def by transfer_prover
end

lemma hom_bind_state:
  assumes "monad_hom m1 m2 h"
  shows "hom_state h (bind_state m1 x f) = bind_state m2 (hom_state h x) (hom_state h \<circ> f)"
proof -
  interpret monad_hom m1 m2 h by fact
  show ?thesis by(rule ST.expand ext)+(simp add: hom_bind split_def o_def)
qed

lemma hom_return_state:
  assumes "monad_hom m1 m2 h"
  shows "hom_state h (return_state m1 x) = return_state m2 x"
proof -
  interpret monad_hom m1 m2 h by fact
  show ?thesis by(rule ST.expand ext)+(simp add: hom_return split_def o_def)
qed

lemma monad_hom_state [simp, intro!, locale_witness]:
  assumes "monad_hom m1 m2 h"
  shows "monad_hom (ST_scheme more1 m1) (ST_scheme more2 m2) (hom_state h)"
proof -
  interpret monad_hom m1 m2 h by fact
  show ?thesis by unfold_locales(simp_all add: hom_return_state hom_bind_state ST_sel assms)
qed

lemma hom_fail_state:
  assumes "monad_fail_hom m1 m2 h"
  shows "hom_state h (fail_state m1) = fail_state m2"
proof -
  interpret monad_fail_hom m1 m2 h by fact
  show ?thesis by(rule ST.expand ext)+(simp add: hom_fail)
qed

lemma monad_fail_hom_state [simp, intro!, locale_witness]:
  assumes "monad_fail_hom m1 m2 h"
  shows "monad_fail_hom (ST_fail_scheme more1 m1) (ST_fail_scheme more2 m2) (hom_state h)"
proof -
  interpret monad_fail_hom m1 m2 h by fact
  show ?thesis by unfold_locales(simp add: hom_fail_state assms ST_fail_sel)
qed


subsection \<open>The reader/environment monad transformer\<close>

datatype ('r, 'm) envT = EnvT (run_env: "'r \<Rightarrow> 'm")
  for rel: rel_envT'

primrec bind_env :: "('a, 'm, 'more) monad_scheme \<Rightarrow> ('r, 'm) envT \<Rightarrow> ('a \<Rightarrow> ('r, 'm) envT) \<Rightarrow> ('r, 'm) envT"
where "bind_env m (EnvT x) f = EnvT (\<lambda>r. bind m (x r) (\<lambda>y. run_env (f y) r))"

definition return_env :: "('a, 'm, 'more) monad_scheme \<Rightarrow> 'a \<Rightarrow> ('r, 'm) envT"
where "return_env m x = EnvT (\<lambda>_. return m x)"

lemma run_bind_env [simp]: "run_env (bind_env m x f) r = bind m (run_env x r) (\<lambda>y. run_env (f y) r)"
by(cases x) simp

lemma run_return_env [simp]: "run_env (return_env m x) r = return m x"
by(simp add: return_env_def)

context
  fixes m :: "('a, 'm, 'more) monad_scheme"
  assumes "monad m"
begin
interpretation monad m by fact

lemma bind_env_assoc [simp]: "bind_env m (bind_env m x f) g = bind_env m x (\<lambda>y. bind_env m (f y) g)"
by(rule envT.expand ext)+(simp add: bind_assoc)

lemma return_bind_env [simp]: "bind_env m (return_env m x) f = f x"
by(rule envT.expand ext)+(simp add: return_bind)

lemma bind_return_env [simp]: "bind_env m x (return_env m) = x"
by(rule envT.expand ext)+(simp add: bind_return)
end

context fixes m :: "('a, 'm, 'more) monad_fail_scheme" begin

definition fail_env :: "('s, 'm) envT"
where "fail_env = EnvT (\<lambda>_. fail m)"

lemma run_fail_env [simp]: "run_env fail_env r = fail m"
by(simp add: fail_env_def)

context assumes "monad_fail m" begin
interpretation monad_fail m by fact

lemma fail_bind_env [simp]: "bind_env m fail_env f = fail_env"
by(rule envT.expand ext)+(simp add: fail_bind)

end
end

context begin interpretation lifting_syntax .

inductive rel_envT :: "('r1 \<Rightarrow> 'r2 \<Rightarrow> bool) \<Rightarrow> ('m1 \<Rightarrow> 'm2 \<Rightarrow> bool) \<Rightarrow> ('r1, 'm1) envT \<Rightarrow> ('r2, 'm2) envT \<Rightarrow> bool"
  for R M x y
where "(R ===> M) (run_env x) (run_env y) \<Longrightarrow> rel_envT R M x y"

inductive_simps rel_envT_EnvT1: "rel_envT R M (EnvT x) y"
inductive_simps rel_envT_EnvT2: "rel_envT R M x (EnvT y)"
inductive_simps rel_envT_simps [simp]: "rel_envT R M (EnvT x) (EnvT y)"

lemma rel_envT_alt_def: "rel_envT S M = BNF_Def.vimage2p run_env run_env (S ===> M)"
apply(rule ext)+
subgoal for x y
  apply(cases x; cases y)
  apply(simp add: rel_ST.simps vimage2p_def)
  done
done

lemma inj_run_env [simp]: "inj run_env"
by(auto simp add: inj_on_def intro: envT.expand)

lemma surj_run_env [simp]: "surj run_env"
by(rule surjI[where f="EnvT"]) simp

lemma rel_envT_eq [relator_eq]: "rel_envT op = op = = op ="
unfolding rel_envT_alt_def rel_fun_eq by (simp add: eq_OO eq_alt[symmetric])

lemma left_unique_rel_envT [transfer_rule]:
  "\<lbrakk> left_total R; left_unique M \<rbrakk> \<Longrightarrow> left_unique (rel_envT R M)"
unfolding rel_envT_alt_def by(simp add: left_unique_fun)

lemma right_unique_rel_envT [transfer_rule]:
  "\<lbrakk> right_total R; right_unique M \<rbrakk> \<Longrightarrow> right_unique (rel_envT R M)"
unfolding rel_envT_alt_def by(simp add: right_unique_fun)

lemma bi_unique_rel_envT [transfer_rule]:
  "\<lbrakk> bi_total R; bi_unique M \<rbrakk> \<Longrightarrow> bi_unique (rel_envT R M)"
unfolding bi_unique_alt_def bi_total_alt_def by(simp add: left_unique_rel_envT right_unique_rel_envT)

lemma right_total_rel_envT [transfer_rule]:
  "\<lbrakk> right_unique R; right_total M \<rbrakk> \<Longrightarrow> right_total (rel_envT R M)"
unfolding rel_envT_alt_def by(simp add: right_total_fun)

lemma left_total_rel_envT [transfer_rule]:
  "\<lbrakk> left_unique R; left_total M \<rbrakk> \<Longrightarrow> left_total (rel_envT R M)"
unfolding rel_envT_alt_def by(simp add: left_total_fun)

lemma bi_total_rel_envT [transfer_rule]:
  "\<lbrakk> bi_unique R; bi_total M \<rbrakk> \<Longrightarrow> bi_total (rel_envT R M)"
unfolding bi_unique_alt_def bi_total_alt_def by(simp add: left_total_rel_envT right_total_rel_envT)

lemma run_env_parametric [transfer_rule]: "(rel_envT R M ===> R ===> M) run_env run_env"
by(rule rel_funI)(erule rel_envT.cases)

lemma envT_parametric [transfer_rule]: "((R ===> M) ===> rel_envT R M) EnvT EnvT"
by(rule rel_funI)(rule rel_envT.intros; simp)

lemma rec_envT_parametric [transfer_rule]:
  "(((R ===> M) ===> C) ===> rel_envT R M ===> C) rec_envT rec_envT"
apply(rule rel_funI)+
subgoal for f g x y
  apply(cases x)
  apply(cases y)
  apply(auto dest: rel_funD)
  done
done

lemma bind_env_parametric [transfer_rule]:
  "(rel_monad_ext A M X ===> rel_envT R M ===> (A ===> rel_envT R M) ===> rel_envT R M) bind_env bind_env"
unfolding bind_env_def[abs_def] by transfer_prover

lemma return_env_parametric [transfer_rule]:
  "(rel_monad_ext A M X ===> A ===> rel_envT R M) return_env return_env"
unfolding return_env_def[abs_def] by transfer_prover

lemma fail_env_parametric [transfer_rule]:
  "(rel_monad_fail_scheme A M X ===> rel_envT R M) fail_env fail_env"
unfolding fail_env_def[abs_def] by transfer_prover

end

definition envT_ext :: "(('a, 'm, 'more) monad_scheme \<Rightarrow> 'more') 
  \<Rightarrow> ('a, 'm, 'more) monad_scheme \<Rightarrow> ('a, ('r, 'm) envT, 'more') monad_scheme"
where
  "envT_ext MORE m = \<lparr>bind = bind_env m, return = return_env m, \<dots> = MORE m\<rparr>"

abbreviation envT_scheme  :: "(('a, 'm, 'more) monad_scheme \<Rightarrow> 'more') 
  \<Rightarrow> ('a, 'm, 'more) monad_scheme \<Rightarrow> ('a, ('r, 'm) envT, 'more') monad_scheme"
where "envT_scheme \<equiv> envT_ext"

abbreviation envT :: "('a, 'm, 'more) monad_scheme \<Rightarrow> ('a, ('r, 'm) envT, 'more) monad_scheme"
where "envT \<equiv> envT_scheme monad.more"

lemma envT_sel:
  shows bind_envT: "bind (envT_scheme MORE m) = bind_env m"
  and return_envT: "return (envT_scheme MORE m) = return_env m"
  and more_envT [simp]: "monad.more (envT_scheme MORE m) = MORE m"
by(simp_all add: envT_ext_def)

lemma monad_lift_envT [simp]:
  "\<And>more. monad.lift (envT_scheme more m) f (EnvT x) = EnvT (monad.lift m f \<circ> x)"
by(simp add: monad.lift_def envT_sel o_def)

lemma run_bind_envT [simp]:
  "\<And>more. run_env (bind (envT_scheme more m) x f) r = 
  bind m (run_env x r) (\<lambda>a. run_env (f a) r)"
by(simp add: bind_envT)

lemma run_return_envT [simp]: "\<And>more. run_env (return (envT_scheme more m) x) r = return m x"
by(simp add: return_envT)

lemma run_lift_envT [simp]:
  "\<And>more. run_env (monad.lift (envT_scheme more m) f x) r = monad.lift m f (run_env x r)"
by(cases x) simp

lemma monad_envT [simp, intro!, locale_witness]: "\<And>more. monad m \<Longrightarrow> monad (envT_scheme more m)"
by unfold_locales(simp_all add: envT_sel)

context begin interpretation lifting_syntax .
lemma envT_ext_parametric [transfer_rule]:
  "((rel_monad_ext A M X ===> X') ===> rel_monad_ext A M X 
  ===> rel_monad_ext A (rel_envT R M) X')
    envT_ext envT_ext"
unfolding envT_ext_def[abs_def] by transfer_prover
end

definition envT_fail_ext :: "(('a, 'm, 'more) monad_fail_scheme \<Rightarrow> 'more')
  \<Rightarrow> ('a, 'm, 'more) monad_fail_scheme \<Rightarrow> (('r, 'm) envT, 'more') monad_fail_ext"
where "envT_fail_ext MORE m = \<lparr>fail = fail_env m, \<dots> = MORE m\<rparr>"

abbreviation envT_fail_scheme :: "(('a, 'm, 'more) monad_fail_scheme \<Rightarrow> 'more')
  \<Rightarrow> ('a, 'm, 'more) monad_fail_scheme \<Rightarrow> ('a, ('r, 'm) envT, 'more') monad_fail_scheme"
where "envT_fail_scheme MORE \<equiv> envT_scheme (envT_fail_ext MORE)"

abbreviation envT_fail :: "('a, 'm, 'more) monad_fail_scheme \<Rightarrow> ('a, ('r, 'm) envT, 'more) monad_fail_scheme"
where "envT_fail \<equiv> envT_fail_scheme monad_fail.more"

lemma envT_fail_sel:
  shows fail_envT_fail: "\<And>more. fail (envT_fail_scheme more m) = fail_env m"
  and more_envT_fail [simp]: "\<And>more. monad_fail.more (envT_fail_scheme more m) = more m"
by(simp_all add: envT_fail_ext_def envT_ext_def)

lemma run_fail_envT [simp]: "\<And>more. run_env (fail (envT_fail_scheme more m)) s = fail m"
by(simp add: fail_envT_fail)

lemma monad_fail_envT [simp, intro!, locale_witness]:
  fixes more m
  assumes "monad_fail m"
  shows "monad_fail (envT_fail_scheme more m)"
proof -
  interpret monad_fail m by fact
  show ?thesis by unfold_locales(rule envT.expand, simp add: fail_bind fun_eq_iff)
qed

lemma run_bind_option_envT [simp]:
  "\<And>more. run_env (monad.bind_option (envT_fail_scheme more m) x f) s = 
  monad.bind_option m x (\<lambda>y. run_env (f y) s)"
by(cases x) simp_all

subsubsection \<open> Transformation of homomorphisms \<close>

primrec hom_env :: "('m1 \<Rightarrow> 'm2) \<Rightarrow> ('r, 'm1) envT \<Rightarrow> ('r, 'm2) envT"
where "hom_env h (EnvT x) = EnvT (\<lambda>r. h (x r))"

lemma run_hom_env [simp]: "run_env (hom_env h x) s = h (run_env x s)"
by(cases x)(simp)

context begin interpretation lifting_syntax .
lemma hom_env_parametric[transfer_rule]:
  "((M1 ===> M2) ===> rel_envT S M1 ===> rel_envT S M2) hom_env hom_env"
unfolding hom_env_def by transfer_prover
end

lemma hom_bind_env:
  assumes "monad_hom m1 m2 h"
  shows "hom_env h (bind_env m1 x f) = bind_env m2 (hom_env h x) (hom_env h \<circ> f)"
proof -
  interpret monad_hom m1 m2 h by fact
  show ?thesis by(rule envT.expand ext)+(simp add: hom_bind split_def o_def)
qed

lemma hom_return_env:
  assumes "monad_hom m1 m2 h"
  shows "hom_env h (return_env m1 x) = return_env m2 x"
proof -
  interpret monad_hom m1 m2 h by fact
  show ?thesis by(rule envT.expand ext)+(simp add: hom_return split_def o_def)
qed

lemma monad_hom_env [simp, intro!, locale_witness]:
  assumes "monad_hom m1 m2 h"
  shows "monad_hom (envT_scheme more1 m1) (envT_scheme more2 m2) (hom_env h)"
proof -
  interpret monad_hom m1 m2 h by fact
  show ?thesis by unfold_locales(simp_all add: hom_return_env hom_bind_env envT_sel assms)
qed

lemma hom_fail_env:
  assumes "monad_fail_hom m1 m2 h"
  shows "hom_env h (fail_env m1) = fail_env m2"
proof -
  interpret monad_fail_hom m1 m2 h by fact
  show ?thesis by(rule envT.expand ext)+(simp add: hom_fail)
qed

lemma monad_fail_hom_env [simp, intro!, locale_witness]:
  assumes "monad_fail_hom m1 m2 h"
  shows "monad_fail_hom (envT_fail_scheme more1 m1) (envT_fail_scheme more2 m2) (hom_env h)"
proof -
  interpret monad_fail_hom m1 m2 h by fact
  show ?thesis by unfold_locales(simp add: hom_fail_env assms envT_fail_sel)
qed

end