section {* Negligibility *}

theory Negligible imports
  Complex_Main
  "~~/src/HOL/Library/Function_Algebras"
  "~~/src/HOL/Library/Polynomial"
begin

lemma less_power2: 
  assumes "2 \<le> a"
  shows "n < a ^ n"
proof -
  have "n < 2 ^ n" by(induct n) simp_all
  also have "\<dots> \<le> a ^ n"
    by(rule power_mono)(simp_all add: assms)
  finally show ?thesis .
qed

lemma le_power2_minus_1:
  "n \<le> 2 ^ n - 1"
proof(induct n)
  case (Suc n)
  hence "Suc n \<le> 2 ^ n" by(cases n) auto
  also have "\<dots> \<le> 2 ^ Suc n - 1" by(cases n) auto
  finally show ?case .
qed simp

lemma less_powr2: 
  fixes x :: real
  assumes "2 \<le> a" -- \<open> probably a bound around 1.44467 would also work \<close>
  shows "x < a powr x"
proof(cases "x \<ge> 0")
  case False
  hence "x < 0" by simp
  also have "a powr x \<ge> 0" by simp
  finally show ?thesis .
next
  case True
  have "x < real (nat (floor x)) + 1" by linarith
  also from le_power2_minus_1[of "nat (floor x)"]
  have "\<dots> \<le> real (2 ^ nat (floor x) :: nat)"
    unfolding of_nat_1[symmetric, where ?'a=real] of_nat_add[symmetric] of_nat_le_iff
    by simp (metis Suc_leI le_refl less_power2)
  also have "\<dots> \<le> 2 powr (nat (floor x))" by(simp add: powr_realpow)
  also have "\<dots> \<le> 2 powr x" by(rule powr_mono)(simp_all add: True)
  also have "\<dots> \<le> a powr x"
    by(rule powr_mono2)(simp_all add: True assms)
  finally show ?thesis .
qed

lemma natceiling_gt_0 [simp]: "nat (ceiling x) > 0 \<longleftrightarrow> x > 0"
by simp

lemma choose_2: "(x choose 2) = x * (x - 1) div 2"
proof(cases "x \<ge> 2")
  case True
  then have "(x choose 2) = fact x div (fact 2 * fact (x - 2))"
    by(rule binomial_altdef_nat)
  also have "fact 2 = 2" by(simp add: eval_nat_numeral)
  also have "fact x = x * (x - 1) * fact (x - 2)"
    using True by(simp add: fact_reduce eval_nat_numeral)
  also have "\<dots> div (2 * fact (x - 2)) = x * (x - 1) div 2"
    by(subst (3) mult.commute)(simp add: div_mult2_eq)
  finally show ?thesis .
next
  case False
  hence "x = 0 \<or> x = 1" by auto
  thus ?thesis by auto
qed

lemma numeral_le_nat [simp]: "numeral n \<le> nat x \<longleftrightarrow> numeral n \<le> x"
by(cases "0 \<le> x")(auto simp add: le_nat_iff not_le elim: less_le_trans)

lemma ceiling_add_one_left [simp]: "\<lceil>1 + x\<rceil> = 1 + \<lceil>x\<rceil>"
by(subst add.commute) simp

lemma powr_fast_growth:
  fixes k :: real
  assumes "k > 1"
  shows "\<exists>n. \<forall>x > n. x < k powr x"
proof -
  def d \<equiv> "k - 1"
  from assms have "0 < d" by(simp add: d_def)
  hence [simp]: "- 1 < 2 / d\<^sup>2"
    by(simp add: field_simps)(metis le_less_trans mult_eq_0_iff power2_eq_square real_minus_mult_self_le zero_less_numeral)
  def n \<equiv> "nat (ceiling (1 + 2 / d ^ 2))"
  have "n \<ge> 2" using \<open>0 < d\<close> by(simp add: n_def)
  { fix x :: real
    assume "x > n"
    with \<open>n \<ge> 2\<close> have "x \<ge> 2" and "x \<ge> 0" by simp_all
    have "x \<le> 1 + real (nat \<lfloor>x\<rfloor>) * 1" by linarith
    also have "2 \<le> real_of_int \<lceil>2 / d\<^sup>2\<rceil> * d\<^sup>2"
      using le_of_int_ceiling[of "2 / d\<^sup>2"] \<open>0 < d\<close> by(simp add: field_simps del: le_of_int_ceiling)
    then have "1 + real (nat \<lfloor>x\<rfloor>) * 1 \<le> 1 + nat \<lfloor>x\<rfloor> * ((n - 1) / 2 * d\<^sup>2)"
      by(intro add_left_mono mult_left_mono)(simp_all add: n_def Suc_nat_eq_nat_zadd1[symmetric])
    also have "\<dots> \<le> 1 + nat \<lfloor>x\<rfloor> * ((nat \<lfloor>x\<rfloor> - 1) / 2 * d\<^sup>2)" using `x > n`
      by(intro add_left_mono mult_left_mono mult_right_mono divide_right_mono)(simp_all, linarith)
    also have "\<dots> = 1 + (nat \<lfloor>x\<rfloor> * (nat \<lfloor>x\<rfloor> - 1) div 2) * d ^ 2"
      by(subst real_of_nat_div) auto
    also have "\<dots> < (\<Sum>k\<in>{0,1,2}. (nat \<lfloor>x\<rfloor> choose k) * d ^ k)"
      using \<open>d > 0\<close> \<open>x \<ge> 2\<close> by(simp add: choose_2)
    also have "\<dots> \<le> (\<Sum>k = 0..nat \<lfloor>x\<rfloor>. (nat \<lfloor>x\<rfloor> choose k) * d ^ k)"
      apply(rule setsum_mono2) using `0 < d` \<open>x \<ge> 2\<close> by(simp_all, linarith)
    also have "\<dots> = (d + 1) ^ nat \<lfloor>x\<rfloor>" by(simp add: binomial_ring)
    also have "\<dots> \<le> k powr (nat \<lfloor>x\<rfloor>)" unfolding d_def using \<open>k > 1\<close> by(simp add: powr_realpow)
    also have "\<dots> \<le> k powr x" using \<open>0 \<le> x\<close> \<open>k > 1\<close> by(intro powr_mono) simp_all
    finally have "x < k powr x" . }
  thus ?thesis by blast
qed

corollary power_fast_growth:
  fixes k :: real
  assumes "k > 1"
  shows "\<exists>n. \<forall>x > n. x < k ^ x"
proof -
  from powr_fast_growth[OF assms] obtain n where n: "\<And>x. x > n \<Longrightarrow> x < k powr x" by blast
  { fix x :: nat
    assume "x > nat \<lceil>n\<rceil>"
    hence "x > n" by linarith 
    hence "x < k powr x" by(rule n)
    also have "\<dots> = k ^ x" using \<open>k > 1\<close> by(simp add: powr_realpow)
    finally have "x < k ^ x" . }
  thus ?thesis by blast
qed

lemma powr_fast_growth2:
  fixes k :: real
  assumes "k > 1" "c > 0"
  shows "\<exists>n. \<forall>x > n. x powr c < k powr x"
proof -
  def k' \<equiv> "k powr (1 / c)"
  have "k' > 1" using assms unfolding k'_def
    by (meson ge_one_powr_ge_zero le_less_trans less_eq_real_def powr_less_mono zero_less_divide_1_iff) 
  from powr_fast_growth[OF this] obtain n where n: "\<And>x. x > n \<Longrightarrow> x < k' powr x" by blast
  { fix x :: real
    assume "x > max 0 n"
    hence "x > 0" "x > n" by simp_all
    from `x > n` have "x < k' powr x" by(rule n)
    with \<open>0 < c\<close> \<open>0 < x\<close> have "x powr c < (k' powr x) powr c"  by(simp add: powr_less_mono2)
    also have "\<dots> = k powr x" unfolding k'_def using \<open>0 < c\<close> by(simp add: powr_powr)
    finally have "x powr c < \<dots>" . }
  thus ?thesis by blast
qed



instantiation "fun" :: (type, abs) abs begin
definition "\<bar>f\<bar> = (\<lambda>x. \<bar>f x\<bar>)"
instance ..
end

lemma abs_fun_apply [simp]: "\<bar>f\<bar> x = \<bar>f x\<bar>"
by(simp add: abs_fun_def)

instance "fun" :: (type, ordered_ab_group_add_abs) ordered_ab_group_add_abs
by intro_classes(auto 4 4 intro: le_funI intro!: abs_leI simp add: abs_ge_self abs_triangle_ineq dest: le_funD)

instance "fun" :: (type, ordered_ring_abs) ordered_ring_abs
by intro_classes(auto 4 3 simp add: fun_eq_iff dest: le_funD intro: abs_eq_mult)

instantiation "fun" :: (type, scaleR) scaleR begin
definition "a *\<^sub>R f = (\<lambda>x. a *\<^sub>R f x)"
instance ..
end

lemma scaleR_fun_apply [simp]: "(a *\<^sub>R f) x = a *\<^sub>R f x"
by(simp add: scaleR_fun_def)

instance "fun" :: (type, real_vector) real_vector
by intro_classes(auto simp add: fun_eq_iff scaleR_right_distrib scaleR_left_distrib)

instance "fun" :: (type, real_algebra) real_algebra
by intro_classes(simp_all add: fun_eq_iff)

instance "fun" :: (type, real_algebra_1) real_algebra_1 ..

definition powr_fun :: "('a \<Rightarrow> real) \<Rightarrow> real \<Rightarrow> 'a \<Rightarrow> real"
where "powr_fun f e = (\<lambda>x. if f x \<le> 0 then 0 else f x powr e)"

locale powr_syntax begin
notation powr_fun (infixr "powr" 80)
end

context begin interpretation powr_syntax .

lemma powr_fun_apply: "(f powr e) x = (if f x \<le> 0 then 0 else f x powr e)"
by(simp add: powr_fun_def)

lemma [simp]:
  shows powr_fun_apply_le0: "f x \<le> 0 \<Longrightarrow> (f powr e) x = 0"
  and powr_fun_apply_gt0: "f x > 0 \<Longrightarrow> (f powr e) x = f x powr e"
by(simp_all add: powr_fun_apply)

lemma powr_fun_add:
  fixes f :: "'a \<Rightarrow> real"
  shows "f powr (a + b) = f powr a * f powr b"
by(simp add: powr_fun_def fun_eq_iff powr_add)

lemma powr_fun_1 [simp]:
  fixes f :: "'a \<Rightarrow> real"
  shows "f \<ge> 0 \<Longrightarrow> f powr 1 = f"
by(auto simp add: powr_fun_def fun_eq_iff le_fun_def intro: antisym)

end

named_theorems negligible_intros

definition negligible :: "(nat \<Rightarrow> real) \<Rightarrow> bool"
where "negligible f \<longleftrightarrow> (\<forall>c>0. \<exists>n. \<forall>x>n. \<bar>f x\<bar> < 1 / (x powr c))"

lemma negligibleI [intro?]:
  "(\<And>c. c > 0 \<Longrightarrow> \<exists>n. \<forall>x>n. real x powr c * \<bar>f x\<bar> < 1) \<Longrightarrow> negligible f"
unfolding negligible_def by(auto simp add: field_simps)

lemma negligibleE [elim?]:
  assumes "negligible f" "c > 0"
  obtains n where "\<forall>x>n. real x powr c * \<bar>f x\<bar> < 1"
using assms unfolding negligible_def by(auto simp add: field_simps)

lemma negligible_mono: "\<lbrakk> negligible g; \<And>\<eta>. \<bar>f \<eta>\<bar> \<le> \<bar>g \<eta>\<bar> \<rbrakk> \<Longrightarrow> negligible f"
apply(rule negligibleI)
apply(erule (1) negligibleE)
apply(auto 4 3 elim!: le_less_trans[rotated] intro!: mult_left_mono simp add: abs_fun_def le_fun_def)
done

lemma negligible_mono2: "\<lbrakk> negligible g; \<And>\<eta>. f \<eta> \<le> g \<eta>; \<And>\<eta>. 0 \<le> f \<eta> \<rbrakk> \<Longrightarrow> negligible f"
by(erule negligible_mono)(metis abs_of_nonneg order_trans)


lemma negligible_K0 [negligible_intros, simp, intro!]: "negligible (\<lambda>_. 0)"
by(rule negligibleI) simp

lemma negligible_0 [negligible_intros, simp, intro!]: "negligible 0"
by(simp add: zero_fun_def)

lemma negligible_const_iff [simp]: "negligible (\<lambda>_. c) \<longleftrightarrow> c = 0"
proof
  assume "negligible (\<lambda>_. c)"
  show "c = 0"
  proof(rule ccontr)
    assume "c \<noteq> 0"
    have "1 > (0 :: real)" by simp
    with `negligible (\<lambda>_. c)` obtain n :: nat
      where negl [rule_format]: "\<forall>x>n. real x powr 1 * \<bar>c\<bar> < 1" ..

    show False
    proof(cases "\<bar>c\<bar> < 1")
      case True
      let ?x = "max (n + 1) (nat (ceiling (1 / \<bar>c\<bar>)))"
      have "1 = \<bar>c\<bar> * (1 / \<bar>c\<bar>)" using `c \<noteq> 0` by simp
      also have "\<dots> \<le> \<bar>c\<bar> * (nat (ceiling (1 / \<bar>c\<bar>)))" 
        by(rule mult_left_mono)(simp_all add: real_nat_ceiling_ge)
      also have "\<dots> \<le> \<bar>c\<bar> * ?x" by(rule mult_left_mono; simp add: max_def; linarith)
      also have "n < ?x" by(simp add: max_def)
      note negl[OF this]
      ultimately show False by(simp add: field_simps)
    next
      case False
      hence "1 \<le> \<bar>c\<bar> * 1" by simp
      also have "\<dots> \<le> \<bar>c\<bar> * (n + 1)" by(rule mult_left_mono) simp_all
      also have "\<dots> < 1" using negl[of "n + 1"] by(simp add: field_simps)
      finally show False by simp
    qed
  qed
qed simp

lemma not_negligible_1 [simp]: "\<not> negligible 1"
by(simp add: one_fun_def)

lemma negligible_plus [negligible_intros]:
  assumes f: "negligible f" 
  and g: "negligible g"
  shows "negligible (\<lambda>\<eta>. f \<eta> + g \<eta>)"
proof
  fix c :: real
  assume "0 < c"
  let ?c = "1 + c"
  from `0 < c` have "0 < ?c" by simp
  from f `0 < ?c` obtain nf where f [rule_format]: "\<forall>x>nf. real x powr ?c * \<bar>f x\<bar> < 1" ..
  from g `0 < ?c` obtain ng where g [rule_format]: "\<forall>x>ng. real x powr ?c * \<bar>g x\<bar> < 1" ..
  let ?n = "max nf ng + 1"
  { fix x :: nat
    let ?xpc = "real x powr ?c"
    assume "x > ?n"
    hence [simp]: "nf < x" "ng < x" "2 \<le> x" "0 < x"
      by(simp_all add: max_def split: split_if_asm)
    have "2 * real x powr c * \<bar>(f + g) x\<bar> \<le> ?xpc * \<bar>(f + g) x\<bar>"
      unfolding powr_add mult.assoc by(rule mult_right_mono)simp_all
    also have "\<dots> \<le> ?xpc * (\<bar>f x\<bar> + \<bar>g x\<bar>)"
      by(rule mult_left_mono)simp_all
    also have "\<dots> = ?xpc * \<bar>f x\<bar> + ?xpc * \<bar>g x\<bar>" by(simp add: field_simps)
    also have "\<dots> < 1 + 1" by(rule add_strict_mono)(rule f g, simp)+
    finally have "real x powr c * \<bar>f x + g x\<bar> < 1" by simp }
  thus "\<exists>n. \<forall>x>n. real x powr c * \<bar>f x + g x\<bar> < 1" by blast
qed

lemma negligible_uminus [simp]:
  "negligible (\<lambda>\<eta>. - f \<eta>) \<longleftrightarrow> negligible f"
by(simp add: negligible_def)

lemma negligible_uminusI [negligible_intros]:
  "negligible f \<Longrightarrow> negligible (\<lambda>\<eta>. - f \<eta>)"
by simp

lemma negligible_minus [negligible_intros]:
  "\<lbrakk> negligible f; negligible g \<rbrakk> \<Longrightarrow> negligible (\<lambda>\<eta>. f \<eta> - g \<eta>)"
by(auto simp add: uminus_add_conv_diff[symmetric] negligible_plus simp del: uminus_add_conv_diff)

lemma negligible_cmult:
  "negligible (\<lambda>\<eta>. c * f \<eta>) \<longleftrightarrow> negligible f \<or> c = 0"
  (is "?lhs \<longleftrightarrow> ?rhs")
proof
  assume ?lhs
  show ?rhs
  proof(rule disjCI)
    assume c: "c \<noteq> 0"
    show "negligible f"
    proof
      fix c' :: real
      assume "c' > 0"
      let ?c' = "(1 / \<bar>c\<bar>) + c'"
      have "0 < ?c'" using `c' > 0` c
        by(simp add: pos_add_strict)
      with `?lhs` obtain n where "\<forall>x>n. x powr ?c' * \<bar>c * f x\<bar> < 1" ..
      hence f: "\<And>x. x > n \<Longrightarrow> \<bar>c\<bar> * real x powr ?c' * \<bar>f x\<bar> < 1"
        by(auto simp add: field_simps abs_mult)
      let ?n = "n + 1"
      { fix x
        assume "x > ?n"
        hence [simp]: "n < x" "2 \<le> x" by simp_all
        have "1 = \<bar>c\<bar> * (1 / \<bar>c\<bar>)" by(simp add: c)
        also have "\<dots> < \<bar>c\<bar> * real x powr (1 / \<bar>c\<bar>)"
          by(rule mult_strict_left_mono)(simp_all add: less_powr2 c)
        finally have "1 * real x powr c' * \<bar>f x\<bar> \<le> \<bar>c\<bar> * real x powr ?c' * \<bar>f x\<bar>"
          unfolding powr_add mult.assoc[symmetric]
          by(intro mult_right_mono) simp_all
        also note f[OF `x > n`]
        finally have "real x powr c' * \<bar>f x\<bar> < 1" by simp }
      thus "\<exists>n. \<forall>x>n. real x powr c' * \<bar>f x\<bar> < 1" by blast
    qed
  qed
next
  assume ?rhs
  thus ?lhs
  proof
    assume "negligible f"
    show ?thesis
    proof
      fix c' :: real
      assume "0 < c'"
      let ?c = "c' + \<bar>c\<bar>"
      have "?c > 0" using `0 < c'` by simp
      with `negligible f` obtain n where f [rule_format]: "\<forall>x>n. x powr ?c * \<bar>f x\<bar> < 1" ..
      let ?n = "n + nat (ceiling (1 / \<bar>c\<bar>)) + 2"
      { fix x :: nat
        assume "x > ?n"
        hence [simp]: "2 \<le> x" "x > n" by simp_all
        have "real x powr c' * \<bar>c * f x\<bar> = \<bar>c\<bar> * (real x powr c' * \<bar>f x\<bar>)" by(simp add: abs_mult)
        also have "\<dots> \<le> real x powr \<bar>c\<bar> * real x powr c' * \<bar>f x\<bar>" unfolding mult.assoc
          by(rule mult_right_mono order.strict_implies_order)+(simp_all add: less_powr2)
        also have "\<dots> = real x powr ?c * \<bar>f x\<bar>" by(simp add: powr_add)
        also note f[OF `x > n`]
        also note calculation }
      thus "\<exists>n. \<forall>x>n. real x powr c' * \<bar>c * f x\<bar> < 1" by blast
    qed
  qed simp
qed

lemma negligible_cmultI [negligible_intros]:
  "(c \<noteq> 0 \<Longrightarrow> negligible f) \<Longrightarrow> negligible (\<lambda>\<eta>. c * f \<eta>)"
by(auto simp add: negligible_cmult)

lemma negligible_multc:
  "negligible (\<lambda>\<eta>. f \<eta> * c) \<longleftrightarrow> negligible f \<or> c = 0"
by(subst mult.commute)(simp add: negligible_cmult)

lemma negligible_multcI [negligible_intros]:
  "(c \<noteq> 0 \<Longrightarrow> negligible f) \<Longrightarrow> negligible (\<lambda>\<eta>. f \<eta> * c)"
by(auto simp add: negligible_multc)

lemma negligible_times [negligible_intros]:
  assumes f: "negligible f"
  and g: "negligible g"
  shows "negligible (\<lambda>\<eta>. f \<eta> * g \<eta>)"
proof
  fix c :: real
  assume "0 < c"
  from f `0 < c` obtain nf where f [rule_format]: "\<forall>x>nf. real x powr c * \<bar>f x\<bar> < 1" ..
  from g `0 < c` obtain ng where g [rule_format]: "\<forall>x>ng. real x powr c * \<bar>g x\<bar> < 1" ..
  let ?n = "max nf ng"
  { fix x
    assume "x > ?n"
    hence [simp]: "nf < x" "ng < x" "1 \<le> real x"
      by(simp_all add: max_def split: split_if_asm)
    have "real x powr c * \<bar>(f * g) x\<bar> \<le> real x powr (c + c) * \<bar>(f * g) x\<bar>"
      by(rule mult_right_mono powr_mono)+(insert `0 < c`, simp_all)
    also have "\<dots> = (real x powr c * \<bar>f x\<bar>) * (real x powr c * \<bar>g x\<bar>)"
      using [[simproc del: field_combine_numerals int_combine_numerals]]
      by(simp add: abs_mult powr_add)
    also have "\<dots> < 1 * 1" by(rule mult_strict_mono)(simp_all add: f g)
    finally have "real x powr c * \<bar>f x * g x\<bar> < 1" by simp }
  thus "\<exists>n. \<forall>x>n. real x powr c * \<bar>f x * g x\<bar> < 1" by blast
qed

lemma negligible_power [negligible_intros]:
  assumes "negligible f"
  and "n > 0"
  shows "negligible (\<lambda>\<eta>. f \<eta> ^ n)"
using `n > 0`
proof(induct n)
  case (Suc n)
  thus ?case using `negligible f`
    by(cases n)(simp_all add: negligible_times)
qed simp

context begin interpretation powr_syntax .

lemma negligible_powr [negligible_intros]:
  assumes "negligible f"
  and "e > 0"
  shows "negligible (f powr e)"
proof
  fix c :: real
  let ?c = "c / e"
  assume "0 < c"
  with \<open>e > 0\<close> have "0 < ?c" by simp
  with \<open>negligible f\<close> obtain n where n[rule_format]: "\<forall>x>n. real x powr ?c * \<bar>f x\<bar> < 1" ..
  { fix x
    assume "x > n"
    have "real x powr c * \<bar>(f powr e) x\<bar> < 1"
    proof(cases "f x \<le> 0")
      case False
      hence "f x > 0" by simp
      with \<open>x > n\<close> \<open>e > 0\<close>
      have "real x powr c * \<bar>(f powr e) x\<bar> = (real x powr ?c * \<bar>f x\<bar>) powr e"
        by(simp add: powr_mult powr_powr)
      also have "0 \<le> real x powr ?c * \<bar>f x\<bar>" using \<open>f x > 0\<close> \<open>x > n\<close> by simp
      with \<open>e > 0\<close> have "\<dots> powr e < 1 powr e" using n
        by(rule powr_less_mono2)(fact \<open>x > n\<close>)
      also have "\<dots> = 1" by simp
      finally show ?thesis .
    qed simp }
  thus "\<exists>n. \<forall>x>n. real x powr c * \<bar>(f powr e) x\<bar> < 1" by blast
qed

end

lemma negligible_times_real [simp]:
  "negligible (\<lambda>\<eta>. f \<eta> * real \<eta>) \<longleftrightarrow> negligible f" (is "?lhs \<longleftrightarrow> ?rhs")
proof
  assume ?lhs
  show ?rhs
  proof
    fix c :: real
    assume "0 < c"
    with \<open>?lhs\<close> obtain n where n[rule_format]: "\<forall>x>n. real x powr c * \<bar>f x * real x\<bar> < 1" ..
    { fix x
      assume "x > n"
      hence "1 \<le> \<bar>real x\<bar>" by simp
      hence "real x powr c * \<bar>f x\<bar> * 1 \<le> real x powr c * \<bar>f x\<bar> * \<bar>real x\<bar>"
        by(rule mult_left_mono) simp
      also have "\<dots> < 1" using n[of x] \<open>x > n\<close> by(simp add: abs_mult)
      finally have "real x powr c * \<bar>f x\<bar> < 1" by simp }
    thus "\<exists>n. \<forall>x>n. real x powr c * \<bar>f x\<bar> < 1" by blast
  qed
next
  assume ?rhs
  show ?lhs
  proof
    fix c :: real
    let ?c = "c + 1"
    assume "0 < c"
    hence "0 < ?c" by simp
    with \<open>?rhs\<close> obtain n where n[rule_format]: "\<forall>x>n. real x powr ?c * \<bar>f x\<bar> < 1" ..
    { fix x
      assume "x > n"
      hence "real x > 0" by simp
      hence "real x powr c * \<bar>f x * real x\<bar> < 1"
        using n[OF \<open>x > n\<close>] by(simp add: abs_mult powr_add algebra_simps) }
    thus "\<exists>n. \<forall>x>n. real x powr c * \<bar>f x * real x\<bar> < 1" by blast
  qed
qed

lemma negligible_times_realI [negligible_intros]: "negligible f \<Longrightarrow> negligible (\<lambda>\<eta>. f \<eta> * real \<eta>)"
by(simp)

lemma negligible_abs [simp]: "negligible (\<lambda>x. \<bar>f x\<bar>) \<longleftrightarrow> negligible f"
by(simp add: negligible_def)

lemma negligible_absI [negligible_intros]: "negligible f \<Longrightarrow> negligible (\<lambda>x. \<bar>f x\<bar>)"
by(simp)

lemma negligible_if_finite:
  assumes "negligible f"
  and "finite {n. P n}" 
  shows "negligible (\<lambda>n. if P n then t n else f n)"
proof
  fix c :: real
  assume "c > 0"
  with \<open>negligible f\<close> obtain n where n: "\<And>x. x > n \<Longrightarrow> x powr c * \<bar>f x\<bar> < 1"
    by(rule negligibleE) blast
  let ?A = "{n. P n}"
  let ?n = "max n (Max (insert 0 ?A))"
  let ?f = "(\<lambda>n. if P n then t n else f n)"
  { fix x
    assume "x > ?n"
    hence "x > n" by simp
    hence "x powr c * \<bar>f x\<bar> < 1" by(rule n)
    moreover have "\<not> P x" using \<open>x > ?n\<close> \<open>finite ?A\<close> by auto
    ultimately have "x powr c * \<bar>?f x\<bar> < 1" by simp }
  thus "\<exists>n. \<forall>x>n. x powr c * \<bar>?f x\<bar> < 1" by blast
qed

lemma negligible_powerI [negligible_intros]:
  assumes "\<bar>k\<bar> < 1"
  shows "negligible (\<lambda>n. k ^ n)"
proof(cases "k = 0")
  case True
  thus ?thesis by(simp add: power_0_left[abs_def] negligible_if_finite)
next
  case False
  hence "0 < \<bar>k\<bar>" by auto
  show ?thesis
  proof
    fix c :: real
    assume "0 < c"
    let ?k = "1 / \<bar>k\<bar>"
    from assms have "?k > 1" using False by(simp add: field_simps)
    from powr_fast_growth2[OF this \<open>0 < c\<close>]
    obtain n where n: "\<And>x. x > n \<Longrightarrow> x powr c < ?k powr x" by blast
    let ?n = "nat \<lceil>n\<rceil>"
    { fix x
      assume "x > ?n"
      hence "x > n" by linarith 
      with n[OF this] \<open>0 < \<bar>k\<bar>\<close> have "x powr c * \<bar>k ^ x\<bar> < 1"
        by(simp add: field_simps power_abs powr_realpow powr_divide) }
    then show "\<exists>n. \<forall>x>n. real x powr c * \<bar>k ^ x\<bar> < 1" by blast
  qed
qed

lemma negligible_inverse_powerI [negligible_intros]: "\<bar>k\<bar> > 1 \<Longrightarrow> negligible (\<lambda>\<eta>. 1 / k ^ \<eta>)"
using negligible_powerI[of "1 / k"] by(simp add: power_one_over)


inductive polynomial :: "('a \<Rightarrow> 'a :: comm_semiring_1) \<Rightarrow> bool"
where
  one [simp, intro!]: "polynomial 1"
| id [simp, intro!]: "polynomial id"
| scalar: "polynomial f \<Longrightarrow> polynomial ((\<lambda>_. c) * f)"
| plus: "\<lbrakk> polynomial f; polynomial g \<rbrakk> \<Longrightarrow> polynomial (f + g)"
| times: "\<lbrakk> polynomial f; polynomial g \<rbrakk> \<Longrightarrow> polynomial (f * g)"

lemmas polynomial_1 = one
  and polynomial_id = id
  and polynomial_scalarI = scalar
  and polynomial_plusI = plus
  and polynomial_timesI = times

hide_fact (open) one id scalar plus times

lemma polynomial_id' [simp, intro!]: "polynomial (\<lambda>x. x)"
by(simp add: id_def[symmetric])

lemma polynomial_K [simp, intro!]: "polynomial (\<lambda>_. c)"
using polynomial_scalarI[where c=c and f=1] by simp

lemma polynomial_uminus [iff]:
  fixes f :: "'a :: comm_ring_1 \<Rightarrow> 'a"
  shows "polynomial (- f) \<longleftrightarrow> polynomial f"
using polynomial_scalarI[where c="- 1" and f=f] polynomial_scalarI[where c="- 1" and f="- f"]
by(auto simp add: fun_Compl_def times_fun_def)

lemma polynomial_scalar [simp]:
  fixes f :: "'a :: {field} \<Rightarrow> 'a"
  shows "polynomial ((\<lambda>_. c) * f) \<longleftrightarrow> c = 0 \<or> polynomial f"
apply(cases "c = 0")
 apply(simp add: polynomial_scalarI times_fun_def)
using polynomial_scalarI[where c="1 / c" and f="(\<lambda>_. c) * f"] polynomial_scalarI[where c=c and f=f]
apply(auto simp add: field_simps times_fun_def)
done

lemma polynomial_numeral [simp, intro!]: "polynomial (numeral n)"
by(induct n)(simp_all only: numeral.simps polynomial_plusI polynomial_1)

lemma polynomial_0 [simp, intro!]: "polynomial 0"
by(simp add: zero_fun_def)

lemma polynomial_minusI:
  fixes f g :: "_ :: comm_ring_1 \<Rightarrow> _"
  shows "\<lbrakk> polynomial f; polynomial g \<rbrakk> \<Longrightarrow> polynomial (f - g)"
by (metis polynomial.simps polynomial_uminus uminus_add_conv_diff)

lemma polynomial_plusD1:
  fixes f g :: "_ :: comm_ring_1 \<Rightarrow> _"
  assumes "polynomial (f + g)" "polynomial f"
  shows "polynomial g"
using polynomial_minusI[OF assms] by simp

lemma polynomial_plusD2:
  fixes f g :: "_ :: comm_ring_1 \<Rightarrow> _"
  assumes "polynomial (f + g)" "polynomial g"
  shows "polynomial f"
using polynomial_minusI[OF assms] by simp

(*
lemma polynomial_timesD1:
  assumes "polynomial (f * g)" "polynomial f"
  shows "polynomial g"
proof -
  def h \<equiv> "f * g"
  let ?Q = "\<lambda>h g :: 'a \<Rightarrow> 'a. \<forall>f. h = f * g \<longrightarrow> polynomial f"
  let ?P = "\<lambda>h :: 'a \<Rightarrow> 'a. \<forall>g. polynomial g \<longrightarrow> ?Q h g"
  from `polynomial (f * g)` h_def have "polynomial h" by simp
  hence "?P h"
  proof(induction rule: polynomial.induct[where P="?P"])
    case one
    show "?P 1"
      apply(rule allI)
      apply(rule impI)
      apply(erule_tac P="?Q 1" in polynomial.induct)
      apply simp
      apply simp
    show ?case
      apply(
    
using assms
apply(
*)

lemma "1_comp" [simp]: "1 \<circ> f = 1"
by(simp add: one_fun_def o_def)

lemma plus_comp_distrib: "f + g \<circ> h = (f \<circ> h) + (g \<circ> h)"
by(simp add: plus_fun_def fun_eq_iff)

lemma times_comp_distrib: "f * g \<circ> h = (f \<circ> h) * (g \<circ> h)"
by(simp add: times_fun_def fun_eq_iff)

lemma K_comp [simp]: "(\<lambda>_. c) \<circ> f = (\<lambda>_. c)"
by(simp add: o_def)

lemma polynomial_compI:
  "\<lbrakk> polynomial f; polynomial g \<rbrakk>
  \<Longrightarrow> polynomial (f \<circ> g)"
by(induction arbitrary: rule: polynomial.induct)(simp_all add: times_comp_distrib plus_comp_distrib polynomial_timesI polynomial_plusI)

lemma polynomial_powerI:
  "polynomial f \<Longrightarrow> polynomial (f ^ n)"
by(induct n)(simp_all add: polynomial_timesI)

lemma polynomial_SucI: "polynomial p \<Longrightarrow> polynomial (\<lambda>n. Suc (p n))"
by(drule polynomial.plus[where g="\<lambda>_. 1"])(simp_all add: plus_fun_def)

lemma negligible_times_polynomial2 [negligible_intros]:
  assumes g: "negligible g"
  and f: "polynomial f"
  shows "negligible (\<lambda>\<eta>. g \<eta> * real (f \<eta>))"
using f g
proof(induction arbitrary: g)
  case one
  thus ?case by simp
next
  case id
  thus ?case
    by(simp only: negligible_times_real o_def id_def)
next
  case (scalar f c)
  from scalar.IH[OF scalar.prems]
  have "negligible (\<lambda>\<eta>. c * (g \<eta> * real (f \<eta>)))"
    unfolding negligible_cmult ..
  also have "\<dots> = g * (real \<circ> (\<lambda>_. c) * f)" by(simp add: o_def algebra_simps times_fun_def)
  finally show ?case by simp
next
  case (plus f f')
  from plus.IH[OF plus.prems] have "negligible (\<lambda>\<eta>. g \<eta> * real (f \<eta>) + g \<eta> * real (f' \<eta>))"
    by(rule negligible_plus)
  thus ?case by(simp add: o_def algebra_simps)
next
  case (times f f')
  interpret powr_syntax .
  let ?g = "\<bar>g\<bar> powr (1 / 2)"
  from times.prems have "negligible \<bar>g\<bar>" by(simp add: abs_fun_def)
  hence "negligible ?g" by(rule negligible_powr) simp
  hence "negligible (\<lambda>\<eta>. ?g \<eta> * real (f \<eta>))" "negligible (\<lambda>\<eta>. ?g \<eta> * real (f' \<eta>))" by(rule times.IH)+
  hence "negligible (\<lambda>\<eta>. (?g \<eta> * real (f \<eta>)) * (?g \<eta> * real (f' \<eta>)))" (is "negligible ?gff'")
    by(rule negligible_times)
  also have "?gff' = (\<lambda>\<eta>. \<bar>g \<eta>\<bar> * real (f \<eta> * f' \<eta>))"
    by(simp add: algebra_simps powr_fun_apply fun_eq_iff powr_add[symmetric])
  also have "negligible \<dots> = negligible (\<lambda>\<eta>. \<bar>\<bar>g \<eta>\<bar> * real (f \<eta> * f' \<eta>)\<bar>)" by simp
  also have "(\<lambda>\<eta>. \<bar>\<bar>g \<eta>\<bar> * real (f \<eta> * f' \<eta>)\<bar>) = (\<lambda>\<eta>. \<bar>g \<eta> * real (f \<eta> * f' \<eta>)\<bar>)" by(simp add: fun_eq_iff abs_mult)
  finally show ?case by(simp only: negligible_abs times_fun_apply)
qed

lemma negligible_times_polynomial1 [negligible_intros]:
  "\<lbrakk> polynomial f; negligible g \<rbrakk> \<Longrightarrow> negligible (\<lambda>\<eta>. real (f \<eta>) * g \<eta>)"
by(subst mult.commute)(rule negligible_times_polynomial2)

lemma negligible_divide_polynomial1 [negligible_intros]:
  "\<lbrakk> polynomial f; negligible (\<lambda>\<eta>. 1 / g \<eta>) \<rbrakk> \<Longrightarrow> negligible (\<lambda>\<eta>. real (f \<eta>) / g \<eta>)"
by(drule (1) negligible_times_polynomial1) simp



lemma poly_pCons': "poly (pCons a p) = (\<lambda>_. a) + id * poly p"
by(simp add: fun_eq_iff)

lemma polynomial_poly [simp]: "polynomial (poly p)"
proof(induct p)
  fix a and p :: "'a poly"
  assume "polynomial (poly p)"
  thus "polynomial (poly (pCons a p))"
    by(simp add: poly_pCons' polynomial_plusI polynomial_timesI)
qed simp

lemma polynomial_ex_poly:
  fixes f :: "'a :: comm_semiring_1 \<Rightarrow> 'a"
  assumes "polynomial f"
  shows "\<exists>p. f = poly p"
using assms
proof(induction)
  case one
  thus "\<exists>p :: 'a poly. 1 = poly p" by(rule exI[where x=1])(auto)
next
  case id
  show "\<exists>p :: 'a poly. id = poly p"
    by(rule exI[where x="monom 1 1"])(simp add: fun_eq_iff poly_monom)
next
  case (scalar f c)
  from scalar.IH obtain p where [simp]: "f = poly p" ..
  show "\<exists>p. ((\<lambda>_. c) * f) = poly p"
    by(rule exI[where x="smult c p"])(simp add: fun_eq_iff)
next
  case (plus f g)
  from plus.IH obtain p q where [simp]: "f = poly p" "g = poly q" by blast
  show "\<exists>p. f + g = poly p"
    by(rule exI[where x="p + q"])(simp add: fun_eq_iff)
next
  case (times f g)
  from times.IH obtain p q where [simp]: "f = poly p" "g = poly q" by blast
  show "\<exists>p. f * g = poly p"
    by(rule exI[where x="p * q"])(simp add: fun_eq_iff)
qed

lemma range_poly_polynomial: "range poly = Collect polynomial"
by(auto dest: polynomial_ex_poly)

lemma inj_poly [simp]: "inj (poly :: 'a :: {idom, ring_char_0} poly \<Rightarrow> _)"
by(simp add: inj_on_def poly_eq_poly_eq_iff)

definition poly_of :: "('a :: {idom, ring_char_0} \<Rightarrow> 'a) \<Rightarrow> 'a poly"
where "poly_of f = (if polynomial f then inv poly f else 0)"

lemma poly_of_0 [simp]: "poly_of 0 = 0"
unfolding poly_of_def by(subst inv_f_eq) auto

lemma poly_of_1 [simp]: "poly_of 1 = 1"
unfolding poly_of_def by(subst inv_f_eq) auto

lemma poly_of_id [simp]: "poly_of id = monom 1 1"
unfolding poly_of_def by(subst inv_f_eq)(auto simp add: poly_monom)

lemma poly_of_scalar [simp]: 
  fixes c :: "'a :: {field, ring_char_0}"
  shows "poly_of ((\<lambda>_. c) * f) = smult c (poly_of f)"
unfolding poly_of_def
apply auto
 apply(simp add: times_fun_def)
 apply(subst inv_f_eq)
 apply(simp_all add: poly_0[abs_def])[3]
apply(drule polynomial_ex_poly)
apply clarsimp
apply(subst inv_f_eq)
apply(simp_all add: fun_eq_iff)
done








term degree





(*
definition degree
where "degree = 
*)







lemma real_all_le [simp]: "(\<forall>x ::real. x \<le> y) \<longleftrightarrow> False"
by(auto intro: exI[where x="y + 1"])

inductive degree_rel :: "(real \<Rightarrow> real) \<Rightarrow> nat \<Rightarrow> bool"
where degree_rel: "(\<And>c. c > 0 \<Longrightarrow> \<exists>x. \<forall>y>x. \<bar>f y\<bar> < c * y ^ n) \<Longrightarrow> degree_rel f n"

lemmas degree_relI = degree_rel

hide_fact (open) degree_rel

lemma degree_relI': "(\<And>c y. \<lbrakk> c > 0; x c < y \<rbrakk> \<Longrightarrow> \<bar>f y\<bar> < c * y ^ n) \<Longrightarrow> degree_rel f n"
by(auto simp add: degree_rel.simps)

lemma degree_relE [consumes 2]:
  assumes "degree_rel f n" "0 < c"
  obtains x where "\<And>y. y > x \<Longrightarrow> \<bar>f y\<bar> < c * y ^ n"
using assms by(auto simp add: degree_rel.simps)

lemma degree_relE' [consumes 2]:
  assumes "degree_rel f n" "0 < c"
  obtains x where "y x > x \<Longrightarrow> \<bar>f (y x)\<bar> < c * (y x) ^ n"
using assms by(auto simp add: degree_rel.simps)

lemma degree_rel_0 [simp]: "degree_rel 0 n"
by(auto simp add: degree_rel.simps zero_less_mult_iff zero_less_power_eq)

lemma degree_rel_mono:
  assumes "degree_rel f m"
  and "m \<le> n"
  shows "degree_rel f n"
proof
  fix c :: real
  assume "0 < c"
  hence [simp]: "0 \<le> c" by simp
  from `degree_rel f m` `0 < c` obtain x
    where m: "\<And>y. y > x \<Longrightarrow> \<bar>f y\<bar> < c * y ^ m" by(rule degree_relE) blast
  { fix y
    assume "y > max 1 x"
    hence [simp]: "1 \<le> y" "x < y" "0 \<le> y" by simp_all
    have "\<bar>f y\<bar> < c * y ^ m * 1" by(simp add: m)
    also have "\<dots> \<le> c * y ^ m * y ^ (n - m)"
      by(rule mult_left_mono)(simp_all add: zero_le_mult_iff)
    also have "\<dots> = c * y ^ n" by(simp add: power_add[symmetric] `m \<le> n`)
    also note calculation }
  thus "\<exists>x. \<forall>y>x. \<bar>f y\<bar> < c * y ^ n" by blast
qed

lemma degree_rel_mono1:
  "\<lbrakk> degree_rel g n; \<bar>f\<bar> \<le> \<bar>g\<bar> \<rbrakk> \<Longrightarrow> degree_rel f n"
by(auto 4 4 simp add: degree_rel.simps le_fun_def intro: order_le_less_trans)

lemma degree_rel_K [simp]:
  assumes "c \<noteq> 0"
  shows "degree_rel (\<lambda>_. c) n \<longleftrightarrow> n > 0" (is "?lhs \<longleftrightarrow> ?rhs")
proof(rule iffI, rule ccontr)
  assume "\<not> ?rhs"
  hence [simp]: "n = 0" by simp
  assume ?lhs
  thus False using assms
    by cases(auto simp add: not_less elim: meta_allE[where x="\<bar>c\<bar>"])
next
  assume ?rhs
  then obtain n' where [simp]: "n = Suc n'" by(cases n) auto
  show ?lhs
    by(rule degree_relI'[where x="\<lambda>c'. max 1 (\<bar>c\<bar> / c')"])(auto simp add: assms mult.assoc[symmetric] divide_less_eq less_divide_eq mult.commute intro: mult_less_le_imp_less[where c="1 :: real", simplified])
qed

lemma degree_rel_id [simp]:
  "degree_rel id n \<longleftrightarrow> n \<ge> 2"
proof
  have "degree_rel id 2"
  proof(rule degree_relI')
    fix c y :: real
    assume "0 < c" "1 / c < y"
    hence [simp]: "0 \<le> c" "c \<noteq> 0" "0 < c" "0 < y" "0 \<le> y"
      by(simp_all add: field_simps)(metis order.strict_trans zero_less_mult_pos zero_less_one le_less_linear mult_pos_neg not_one_less_zero order.strict_trans)+

    have "\<bar>id y\<bar> = c * y * (1 / c)" by simp
    also have "\<dots> < c * y * y" by(rule mult_strict_left_mono)(simp_all add: `1 / c < y` zero_less_mult_iff)
    also have "\<dots> = c * y\<^sup>2" by(simp add: power2_eq_square)
    finally show "\<bar>id y\<bar> < c * y\<^sup>2" .
  qed
  moreover
  assume "2 \<le> n"
  ultimately show "degree_rel id n" by(rule degree_rel_mono)
next
  assume *: "degree_rel id n"
  have "n = 0 \<or> n = 1 \<or> 2 \<le> n" by auto
  thus "2 \<le> n"
  proof(elim disjE)
    assume [simp]: "n = 0"
    have "0 < (1 :: real)" by simp
    with * have False
      by(rule degree_relE'[where y="\<lambda>x. max 1 (x + 1)"])(fastforce simp add: max_def split: split_if_asm)
    thus ?thesis ..
  next
    assume [simp]: "n = 1"
    have "0 < (1 :: real)" by simp
    with * have False
      by(rule degree_relE'[where y="\<lambda>x. x + 1"]) auto
    thus ?thesis ..
  qed
qed

lemma degree_rel_scalar:
  assumes "c \<noteq> 0"
  shows "degree_rel ((\<lambda>_. c) * f) n \<longleftrightarrow> degree_rel f n"
proof -
  def P \<equiv> "\<lambda>c. \<exists>x. \<forall>y>x. \<bar>f y\<bar> < c * y ^ n"
  hence "degree_rel ((\<lambda>_. c) * f) n \<longleftrightarrow> (\<forall>c'>0. P (c' / \<bar>c\<bar>))"
    by(simp add: abs_mult field_simps assms degree_rel.simps)
  also have "\<dots> \<longleftrightarrow> (\<forall>c'>0. P c')"
    by(auto simp add: assms zero_less_mult_iff field_simps elim: allE[where x="c' * \<bar>c\<bar>" for c'])
  finally show ?thesis unfolding P_def by(simp add: degree_rel.simps)
qed

lemma degree_rel_scalar_iff:
  "degree_rel ((\<lambda>_. c) * f) n \<longleftrightarrow> c = 0 \<or> degree_rel f n"
apply(cases "c = 0")
 apply(simp add: times_fun_def zero_fun_def[symmetric])
apply(simp add: degree_rel_scalar)
done

lemma degree_rel_plusI:
  assumes "degree_rel f n"
  and "degree_rel g n"
  shows "degree_rel (f + g) n"
proof
  fix c :: real
  assume "0 < c"
  hence "0 < c / 2" and [simp]: "0 \<le> c" by simp_all
  with assms obtain xf xg
    where f: "\<And>y. y > xf \<Longrightarrow> \<bar>f y\<bar> < c / 2 * y ^ n"
    and g: "\<And>y. y > xg \<Longrightarrow> \<bar>g y\<bar> < c / 2 * y ^ n"
    by(blast elim: degree_rel.cases)
  let ?x = "max 1 (max xf xg)"
  { fix y
    assume "y > ?x"
    hence [simp]: "xf < y" "xg < y" "1 \<le> y" by simp_all
    have "\<bar>(f + g) y\<bar> \<le> \<bar>f y\<bar> + \<bar>g y\<bar>" by simp
    also have "\<dots> < c / 2 * y ^ n + c / 2 * y ^ n"
      by(rule add_strict_mono)(rule f g, simp)+
    also have "\<dots> = c * y ^ n" by simp
    also note calculation }
  thus "\<exists>x. \<forall>y>x. \<bar>(f + g) y\<bar> < c * y ^ n" by blast
qed

lemma degree_rel_plus_max:
  "\<lbrakk> degree_rel f n; degree_rel g m \<rbrakk> \<Longrightarrow> degree_rel (f + g) (max n m)"
by(rule degree_rel_plusI)(erule degree_rel_mono, simp)+

lemma degree_rel_timesI:
  assumes "degree_rel f n" "degree_rel g m"
  shows "degree_rel (f * g) (n + m)"
proof
  fix c :: real
  assume "0 < c"
  hence "0 < sqrt c" by simp
  with assms obtain xf xg
    where f: "\<And>y. y > xf \<Longrightarrow> \<bar>f y\<bar> < sqrt c * y ^ n"
    and g: "\<And>y. y > xg \<Longrightarrow> \<bar>g y\<bar> < sqrt c * y ^ m" by(blast elim: degree_rel.cases)
  let ?x = "max 0 (max xf xg)"
  { fix y
    assume "y > ?x"
    hence [simp]: "xf < y" "xg < y" "y \<noteq> 0" "0 < y" by simp_all
    have "\<bar>(f * g) y\<bar> = \<bar>f y\<bar> * \<bar>g y\<bar>" by(simp add: abs_mult)
    also have "\<dots> < (sqrt c * y ^ n) * (sqrt c * y ^ m)"
      by(rule mult_strict_mono)(rule f g, simp_all add: zero_less_mult_iff `0 < c` zero_less_power_eq)+
    also have "\<dots> = c * y ^ (n + m)" using  `0 < c` by(simp add: power_add real_sqrt_mult[symmetric])
    also note calculation }
  thus "\<exists>x. \<forall>y>x. \<bar>(f * g) y\<bar> < c * y ^ (n + m)" by blast
qed

lemma degree_rel_id_times_id: "degree_rel (id * id) n \<longleftrightarrow> n \<ge> 3"
proof
  assume *: "degree_rel (id * id) n"
  have "n = 0 \<or> n = 1 \<or> n = 2 \<or> 3 \<le> n" by auto
  thus "3 \<le> n"
  proof(elim disjE)
    assume "n = 0"
    assume [simp]: "n = 0"
    have "0 < (1 :: real)" by simp
    with * have False
      apply(rule degree_relE'[where y="\<lambda>x. max 1 (x + 1)"])
      apply(erule meta_impE)
      apply(simp_all add: max_def split: split_if_asm)
      by (metis le_add_same_cancel2 le_less_trans less_irrefl mult_le_cancel_left1 not_one_less_zero)
    thus ?thesis ..
  next
    assume [simp]: "n = 1"
    have "0 < (1 :: real)" by simp
    with * have False
      by(rule degree_relE'[where y="\<lambda>x. max 2 (x + 1)"])(fastforce simp add: max_def split: split_if_asm)
    thus ?thesis ..
  next
    assume [simp]: "n = 2"
    have "0 < (1 :: real)" by simp
    with * have False by(rule degree_relE'[where y="\<lambda>x. x + 1"])(simp add: power2_eq_square)
    thus ?thesis ..
  qed
next
  have "degree_rel (id * id) 3"
  proof(rule degree_relI')
    fix c y :: real
    assume "0 < c" "1 / c < y"
    hence [simp]: "0 \<le> c" "c \<noteq> 0" "0 < c" "0 < y" "0 \<le> y"
      by(simp_all add: field_simps)(metis order.strict_trans zero_less_mult_pos zero_less_one le_less_linear mult_pos_neg not_one_less_zero order.strict_trans)+

    have "\<bar>(id * id) y\<bar> = c * y * y * (1 / c)" by simp
    also have "\<dots> < c * y * y * y" by(rule mult_strict_left_mono)(simp_all add: `1 / c < y` zero_less_mult_iff)
    also have "\<dots> = c * y ^ 3" by(simp add: numeral_3_eq_3)
    finally show "\<bar>(id * id) y\<bar> < c * y ^ 3" .
  qed
  moreover
  assume "3 \<le> n"
  ultimately show "degree_rel (id * id) n" by(rule degree_rel_mono)
qed

lemma polynomial_ex_degree_rel:
  "polynomial f \<Longrightarrow> \<exists>n. degree_rel f n"
by(induction rule: polynomial.induct)(auto simp add: id_def[symmetric] degree_rel_scalar_iff intro: degree_rel_timesI degree_rel_plus_max simp del: times_fun_apply plus_fun_apply)


definition degree :: "(real \<Rightarrow> real) \<Rightarrow> nat"
where "degree f = (LEAST n. degree_rel f n)"

lemma degree_0 [simp]: "degree 0 = 0"
by(auto simp add: degree_def intro: Least_equality)

lemma degree_K [simp]: "c \<noteq> 0 \<Longrightarrow> degree (\<lambda>_. c) = 1"
by(auto simp add: degree_def intro: Least_equality)

lemma degree_1 [simp]: "degree 1 = 1"
by(simp add: one_fun_def)

lemma degree_id [simp]: "degree id = 2"
by(auto simp add: degree_def intro: Least_equality)

lemma degree_scalar: "c \<noteq> 0 \<Longrightarrow> degree ((\<lambda>_. c) * f) = degree f"
by(simp add: degree_def degree_rel_scalar)

lemma degree_scalar_iff:
  "degree ((\<lambda>_. c) * f) = (if c = 0 then 0 else degree f)"
by(simp add: degree_scalar zero_fun_def[symmetric])

lemma degree_id_times_id: "degree (id * id) = 3"
by(auto simp add: degree_def degree_rel_id_times_id intro: Least_equality)

lemma degree_times_id:
  assumes "polynomial g" "g \<noteq> 0"
  shows "degree (id * g) = 1 + degree g"
using assms
proof(induction) 
  case one
  show ?case by(simp add: id_def[symmetric])
next
  case id
  show ?case
    by(simp add: id_def[symmetric] degree_id_times_id del: times_fun_apply)
next
  case (scalar f c)
  from scalar.prems have [simp]: "c \<noteq> 0" "f \<noteq> (\<lambda>_. 0)" by auto
  have "degree (id * ((\<lambda>_. c) * f)) = degree ((\<lambda>_. c) * (id * f))" by(simp add: mult_ac)
  also have "\<dots> = degree (id * f)" by(simp add: degree_scalar)
  also have "\<dots> = 1 + degree f" by(rule scalar.IH) simp
  also have "1 + degree f = 1 + degree ((\<lambda>_. c) * f)" by(simp add: degree_scalar)
  finally show ?case .
next
  case (plus f g)
  have "degree (id * (f + g)) = degree (id * f + id * g)" by(simp add: algebra_simps)
  also have "\<dots> = max (degree (id * f)) (degree (id * g))"
oops



lemma degree_times:
  assumes "polynomial f"
  and "polynomial g"
  shows "degree (f * g) = degree f + degree g - 1"
using assms
proof(induct arbitrary: g)
  case one
  show ?case by simp
next
  case id
oops





definition polysgn :: "(real \<Rightarrow> real) \<Rightarrow> real"
where
  "polysgn f = (if \<exists>n. \<forall>x>n. f x > 0 then 1 else if \<exists>n. \<forall>x>n. f x < 0 then -1 else 0)"

lemma polysgn_K [simp]: "polysgn (\<lambda>_. c) = sgn c"
by(auto simp add: polysgn_def)(rule_tac x="n + 1" in exI, simp)+

lemma [simp]:
  shows polysgn_0: "polysgn 0 = 0"
  and polysgn_1: "polysgn 1 = 1"
by(simp_all add: zero_fun_def one_fun_def)

lemma polysgn_id [simp]: "polysgn id = 1"
by(auto simp add: polysgn_def)

lemma polynomial_polysgn_cases_aux:
  fixes f :: "'a :: {linordered_field, ring_1} \<Rightarrow> 'a"
  assumes "polynomial f"
  shows "(\<exists>n. \<forall>x>n. f x > 0) \<or> (\<exists>n. \<forall>x>n. f x < 0) \<or> f = 0"
using assms
proof(induction)
  case one 
  show ?case by simp
next
  case id
  show ?case by auto
next
  case (scalar f c)
  have "0 < c \<or> c < 0 \<or> c = 0" by(metis less_linear)
  with scalar.IH show ?case
    by(auto 4 4 simp add: not_less mult_le_0_iff mult_less_0_iff)
next
  case (plus f g)
  thus ?case
oops

lemma polysgn_scalar [simp]: "polysgn ((\<lambda>_. c) * f) = sgn c * polysgn f"
oops

lemma "\<lbrakk> polynomial f; polysgn f = 0 \<rbrakk> \<Longrightarrow> f = 0"
apply(induct f rule: polynomial.induct)
apply simp
apply simp
oops




inductive polybounded :: "('a :: linordered_idom \<Rightarrow> 'a) \<Rightarrow> bool" where 
  polybounded: "\<lbrakk> polynomial g; \<bar>f\<bar> \<le> \<bar>g\<bar> \<rbrakk> \<Longrightarrow> polybounded f"

hide_fact (open) polybounded

lemmas polyboundedI = polybounded.polybounded

lemma polynomial_polybounded: "polynomial f \<Longrightarrow> polybounded f"
by(erule polyboundedI) simp

lemma [simp, intro!]:
  shows polybounded_const: "polybounded (\<lambda>_. c)"
  and polybounded_0: "polybounded 0"
  and polybounded_1: "polybounded 1"
  and polybounded_numeral: "polybounded (numeral n)"
  and polybounded_id: "polybounded id"
  and polybounded_id': "polybounded (\<lambda>x. x)"
by(simp_all add: polynomial_polybounded)

lemma polybounded_plus:
  "\<lbrakk> polybounded f; polybounded g \<rbrakk> \<Longrightarrow> polybounded (f + g)"
apply(clarsimp simp add: polybounded.simps)
apply(rule exI conjI)+
apply(erule (1) polynomial_plusI)
oops



end